"""
API for interacting with the file system on Hops (HopsFS).

It is a wrapper around pydoop together with utility functions that are Hops-specific.
"""
import socket
from six import string_types
import shutil
import fnmatch
import os
import errno

from hops import constants
from hops.service_discovery import ServiceDiscovery
import sys
import subprocess

import ntpath
import importlib

import logging
log = logging.getLogger(__name__)

# Compatibility with SageMaker
pydoop_available = True
try:
    import pydoop.hdfs as hdfs
    import pydoop.hdfs.path as path
    import pydoop.hdfs.fs as hdfs_fs
except:
    pydoop_available = False

import re
from xml.dom import minidom

tls_enabled = None
webhdfs_address = None

if pydoop_available: 
    # Replace Pydoop split method to be able to support hopsfs:// schemes
    class _HopsFSPathSplitter(hdfs.path._HdfsPathSplitter):
        @classmethod
        def split(cls, hdfs_path, user):
            if not hdfs_path:
                cls.raise_bad_path(hdfs_path, "empty")
            scheme, netloc, path = cls.parse(hdfs_path)
            if not scheme:
                scheme = "file" if hdfs_fs.default_is_local() else "hdfs"
            if scheme == "hdfs" or scheme == "hopsfs":
                if not path:
                    cls.raise_bad_path(hdfs_path, "path part is empty")
                if ":" in path:
                    cls.raise_bad_path(
                        hdfs_path, "':' not allowed outside netloc part"
                    )
                hostname, port = cls.split_netloc(netloc)
                if not path.startswith("/"):
                    path = "/user/%s/%s" % (user, path)
            elif scheme == "file":
                hostname, port, path = "", 0, netloc + path
            else:
                cls.raise_bad_path(hdfs_path, "unsupported scheme %r" % scheme)
            return hostname, port, path

    hdfs.path._HdfsPathSplitter = _HopsFSPathSplitter

def get_plain_path(abs_path):
    """
    Convert absolute HDFS/HOPSFS path to plain path (dropping prefix and ip)

    Example use-case:

    >>> hdfs.get_plain_path("hdfs://10.0.2.15:8020/Projects/demo_deep_learning_admin000/Models/")
    >>> # returns: "/Projects/demo_deep_learning_admin000/Models/"

     Args:
         :abs_path: the absolute HDFS/hopsfs path containing prefix and/or ip

    Returns:
          the plain path without prefix and ip
    """
    return path.split(path.abspath(abs_path))[2]

def project_id():
    """
    Get the Hopsworks project id from environment variables

    Returns: the Hopsworks project id

    """
    return os.environ[constants.ENV_VARIABLES.HOPSWORKS_PROJECT_ID_ENV_VAR]

def project_user():
    """
    Gets the project username ("project__user") from environment variables

    Returns:
        the project username
    """

    try:
        hops_user = os.environ[constants.ENV_VARIABLES.HADOOP_USER_NAME_ENV_VAR]
    except:
        hops_user = os.environ[constants.ENV_VARIABLES.HDFS_USER_ENV_VAR]
    return hops_user

def project_name():
    """
    Extracts the project name from the project username ("project__user") or from the environment if available

    Returns:
        project name
    """
    try:
        return os.environ[constants.ENV_VARIABLES.HOPSWORKS_PROJECT_NAME_ENV_VAR]
    except:
        pass

    hops_user = project_user()
    hops_user_split = hops_user.split("__")  # project users have username project__user
    project = hops_user_split[0]
    return project

def project_path(project=None, exclude_nn_addr=False):
    """ Get the path in HopsFS where the HopsWorks project is located. To point to a particular dataset, this path should be
    appended with the name of your dataset.

    >>> from hops import hdfs
    >>> project_path = hdfs.project_path()
    >>> print("Project path: {}".format(project_path))

    Args:
        :project: If this value is not specified, it will get the path to your project. If you need to path to another project, you can specify the name of the project as a string.

    Returns:
        returns the project absolute path
    """

    if project is None:
        project = project_name()

    # abspath means "hdfs://namenode:port/ is preprended
    abspath = hdfs.path.abspath("/Projects/" + project + "/")
    if exclude_nn_addr:
        abspath = re.sub(r"\d+.\d+.\d+.\d+:\d+", "", abspath)
    return abspath


def get():
    """ Get a handle to pydoop hdfs using the default namenode (specified in hadoop config)

    Returns:
        Pydoop hdfs handle
    """
    return hdfs.hdfs('default', 0, user=project_user())


def get_fs():
    """ Get a handle to pydoop fs using the default namenode (specified in hadoop config)

    Returns:
        Pydoop fs handle
    """
    return hdfs.fs.hdfs('default', 0, user=project_user())


def _expand_path(hdfs_path, project="", exists=True):
    """
    Expands a given path. If the path is /Projects.. hdfs:// is prepended.
    If the path is ../ the full project path is prepended.

    Args:
        :hdfs_path the path to be expanded
        :exists boolean flag, if this is true an exception is thrown if the expanded path does not exist.

    Raises:
        IOError if exists flag is true and the path does not exist

    Returns:
        path expanded with HDFS and project
    """
    if not isinstance(hdfs_path, string_types):
        hdfs_path = hdfs_path.decode()
    if project == "":
        project = project_name()
    # Check if a full path is supplied. If not, assume it is a relative path for this project - then build its full path and return it.
    if hdfs_path.startswith("/Projects/") or hdfs_path.startswith("/Projects"):
        hdfs_path = "hdfs://" + hdfs_path
    elif not (hdfs_path.startswith("hdfs://") or hdfs_path.startswith("hopsfs://")):
        # if the file URL type is not HDFS, throw an error
        if "://" in hdfs_path:
            raise IOError("path %s must be a full hopsfs path or a relative path" % hdfs_path)
        proj_path = project_path(project)
        hdfs_path = proj_path + hdfs_path
    if exists == True and not hdfs.path.exists(hdfs_path):
        raise IOError("path %s not found" % hdfs_path)
    return hdfs_path

def copy_to_hdfs(local_path, relative_hdfs_path, overwrite=False, project=None):
    """
    Copies a path from local filesystem to HDFS project (recursively) using relative path in $CWD to a path in hdfs (hdfs_path)

    For example, if you execute:

    >>> copy_to_hdfs("data.tfrecords", "/Resources", project="demo")

    This will copy the file data.tfrecords to hdfs://Projects/demo/Resources/data.tfrecords

    Args:
        :local_path: Absolute or local path on the local filesystem to copy
        :relative_hdfs_path: a path in HDFS relative to the project root to where the local path should be written
        :overwrite: a boolean flag whether to overwrite if the path already exists in HDFS
        :project: name of the project, defaults to the current HDFS user's project
    """
    if project == None:
        project = project_name()

    # Absolute path
    if os.path.isabs(local_path):
        full_local = local_path
    else:
        full_local = os.getcwd() + '/' + local_path

    hdfs_path = _expand_path(relative_hdfs_path, project, exists=False)

    if overwrite:
        hdfs_path = hdfs_path + "/" + os.path.basename(full_local)
        if exists(hdfs_path):
            # delete hdfs path since overwrite flag was set to true
            delete(hdfs_path, recursive=True)

    log.debug("Started copying local path {} to hdfs path {}\n".format(local_path, hdfs_path))

    # copy directories from local path to HDFS project path
    hdfs.put(full_local, hdfs_path)

    log.debug("Finished copying\n")


def delete(hdfs_path, recursive=False):
    """
    Deletes path, path can be absolute or relative.
    If recursive is set to True and path is a directory, then files will be deleted recursively.

    For example

    >>> delete("/Resources/", recursive=True)

    will delete all files recursively in the folder "Resources" inside the current project.

    Args:
        :hdfs_path: the path to delete (project-relative or absolute)

    Returns:
        None

    Raises:
        IOError when recursive is False and directory is non-empty
    """
    hdfs_path = _expand_path(hdfs_path)
    hdfs_handle = get()
    if hdfs_handle.exists(hdfs_path):
        hdfs_handle.delete(hdfs_path, recursive=recursive)


def copy_to_local(hdfs_path, local_path="", overwrite=False, project=None):
    """
    Copies a directory or file from a HDFS project to a local private scratch directory. If there is not enough space on the local scratch directory, an exception is thrown.
    If the local file exists, and the hdfs file and the local file are the same size in bytes, return 'ok' immediately.
    If the local directory tree exists, and the hdfs subdirectory and the local subdirectory have the same files and directories, return 'ok' immediately.

    For example, if you execute:

    >>> copy_to_local("Resources/my_data")

    This will copy the directory my_data from the Resources dataset in your project to the current working directory on the path ./my_data

    Raises:
      IOError if there is not enough space to localize the file/directory in HDFS to the scratch directory ($PDIR)

    Args:
        :hdfs_path: You can specify either a full hdfs pathname or a relative one (relative to your Project's path in HDFS).
        :local_path: the relative or full path to a directory on the local filesystem to copy to (relative to a scratch directory $PDIR), defaults to $CWD
        :overwrite: a boolean flag whether to overwrite if the path already exists in the local scratch directory.
        :project: name of the project, defaults to the current HDFS user's project

    Returns:
        the full local pathname of the file/dir
    """

    if project == None:
        project = project_name()

    if os.path.isabs(local_path):
        local_dir = local_path
    else:
        local_dir = os.getcwd() + '/' + local_path

    if not os.path.isdir(local_dir):
        raise IOError("You need to supply the path to a local directory. This is not a local dir: %s" % local_dir)

    filename = path.basename(hdfs_path)
    full_local = local_dir + "/" + filename

    project_hdfs_path = _expand_path(hdfs_path, project=project)

    # Get the amount of free space on the local drive
    stat = os.statvfs(local_dir)
    free_space_bytes = stat.f_bsize * stat.f_bavail

    hdfs_size = path.getsize(project_hdfs_path)

    if os.path.isfile(full_local) and not overwrite:
        sz = os.path.getsize(full_local)
        if hdfs_size == sz:
            log.info("File " + project_hdfs_path + " is already localized, skipping download...")
            return full_local
        else:
            os.remove(full_local)

    if os.path.isdir(full_local) and not overwrite:
        try:
            localized = _is_same_directory(full_local, project_hdfs_path)
            if localized:
                log.info("Full directory subtree already on local disk and unchanged. Set overwrite=True to force download")
                return full_local
            else:
                shutil.rmtree(full_local)
        except Exception as e:
            log.error("Failed while checking directory structure to avoid re-downloading dataset, falling back to downloading")
            log.error(e)
            shutil.rmtree(full_local)

    if hdfs_size > free_space_bytes:
        raise IOError("Not enough local free space available on scratch directory: %s" % local_path)

    if overwrite:
        if os.path.isdir(full_local):
            shutil.rmtree(full_local)
        elif os.path.isfile(full_local):
            os.remove(full_local)

    log.debug("Started copying " + project_hdfs_path + " to local disk on path " + local_dir + "\n")

    hdfs.get(project_hdfs_path, local_dir)

    log.debug("Finished copying\n")

    return full_local


def _is_same_directory(local_path, hdfs_path):
    """
    Validates that the same occurrence and names of files exists in both hdfs and local
    """
    local_file_list = []
    for root, dirnames, filenames in os.walk(local_path):
        for filename in fnmatch.filter(filenames, '*'):
            local_file_list.append(filename)
        for dirname in fnmatch.filter(dirnames, '*'):
            local_file_list.append(dirname)
    local_file_list.sort()

    hdfs_file_list = glob(hdfs_path + '/*', recursive=True)
    hdfs_file_list = [path.basename(str(r)) for r in hdfs_file_list]
    hdfs_file_list.sort()

    if local_file_list == hdfs_file_list:
        return True
    else:
        return False

def cp(src_hdfs_path, dest_hdfs_path, overwrite=False):
    """
    Copy the contents of src_hdfs_path to dest_hdfs_path.

    If src_hdfs_path is a directory, its contents will be copied recursively.
    Source file(s) are opened for reading and copies are opened for writing.

    Args:
        :src_hdfs_path: You can specify either a full hdfs pathname or a relative one (relative to your Project's path in HDFS).
        :dest_hdfs_path: You can specify either a full hdfs pathname or a relative one (relative to your Project's path in HDFS).
        :overwrite: boolean flag whether to overwrite destination path or not.

    """
    src_hdfs_path = _expand_path(src_hdfs_path)
    dest_hdfs_path = _expand_path(dest_hdfs_path, exists=False)

    if overwrite and exists(dest_hdfs_path):
        # delete path since overwrite flag was set to true
        delete(dest_hdfs_path, recursive=True)

    hdfs.cp(src_hdfs_path, dest_hdfs_path)

def glob(hdfs_path, recursive=False, project=None):
    """
    Finds all the pathnames matching a specified pattern according to the rules used by the Unix shell, although results are returned in arbitrary order.

    Globbing gives you the list of files in a dir that matches a supplied pattern

    >>> glob('Resources/*.json')
    >>> ['Resources/1.json', 'Resources/2.json']

    glob is implemented as  os.listdir() and fnmatch.fnmatch()
    We implement glob as hdfs.ls() and fnmatch.filter()

    Args:
     :hdfs_path: You can specify either a full hdfs pathname or a relative one (relative to project_name in HDFS.
     :project: If the supplied hdfs_path is a relative path, it will look for that file in this project's subdir in HDFS.

    Raises:
        IOError if the supplied hdfs path does not exist

    Returns:
      A possibly-empty list of path names that match pathname, which must be a string containing a path specification. pathname can be either absolute
    """

    # Get the full path to the dir for the input glob pattern
    # "hdfs://Projects/jim/blah/*.jpg" => "hdfs://Projects/jim/blah"
    # Then, ls on 'hdfs://Projects/jim/blah', then filter out results
    if project == None:
        project = project_name()
    lastSep = hdfs_path.rfind("/")
    inputDir = hdfs_path[:lastSep]
    inputDir = _expand_path(inputDir, project)
    pattern = hdfs_path[lastSep + 1:]
    if not hdfs.path.exists(inputDir):
        raise IOError("Glob path %s not found" % inputDir)
    dirContents = hdfs.ls(inputDir, recursive=recursive)
    return fnmatch.filter(dirContents, pattern)


def ls(hdfs_path, recursive=False, project=None):
    """
    Returns all the pathnames in the supplied directory.

    Args:
        :hdfs_path: You can specify either a full hdfs pathname or a relative one (relative to project_name in HDFS).
        :recursive: if it is a directory and recursive is True, the list contains one item for every file or directory in the tree rooted at hdfs_path.
        :project: If the supplied hdfs_path is a relative path, it will look for that file in this project's subdir in HDFS.

    Returns:
      A possibly-empty list of path names stored in the supplied path.
    """
    if project == None:
        project = project_name()
    hdfs_path = _expand_path(hdfs_path, project)
    return hdfs.ls(hdfs_path, recursive=recursive)


def lsl(hdfs_path, recursive=False, project=None):
    """
    Returns all the pathnames in the supplied directory.

    Args:
        :hdfs_path: You can specify either a full hdfs pathname or a relative one (relative to project_name in HDFS).
        :recursive: if it is a directory and recursive is True, the list contains one item for every file or directory in the tree rooted at hdfs_path.
        :project: If the supplied hdfs_path is a relative path, it will look for that file in this project's subdir in HDFS.

    Returns:
        A possibly-empty list of path names stored in the supplied path.
    """
    if project == None:
        project = project_name()
    hdfs_path = _expand_path(hdfs_path, project)
    return hdfs.lsl(hdfs_path, recursive=recursive)


def rmr(hdfs_path, project=None):
    """
    Recursively remove files and directories.

    Args:
        :hdfs_path: You can specify either a full hdfs pathname or a relative one (relative to project_name in HDFS).
        :project: If the supplied hdfs_path is a relative path, it will look for that file in this project's subdir in HDFS.

    """
    if project == None:
        project = project_name()
    hdfs_path = _expand_path(hdfs_path, project)
    return hdfs.rmr(hdfs_path)


def mkdir(hdfs_path, project=None):
    """
    Create a directory and its parents as needed.

    Args:
        :hdfs_path: You can specify either a full hdfs pathname or a relative one (relative to project_name in HDFS).
        :project: If the supplied hdfs_path is a relative path, it will look for that file in this project's subdir in HDFS.

    """
    if project == None:
        project = project_name()
    hdfs_path = _expand_path(hdfs_path, project, exists=False)
    return hdfs.mkdir(hdfs_path)


def move(src, dest):
    """
    Move or rename src to dest.

    Args:
        :src: You can specify either a full hdfs pathname or a relative one (relative to your Project's path in HDFS).
        :dest: You can specify either a full hdfs pathname or a relative one (relative to your Project's path in HDFS).

    """
    src = _expand_path(src, project_name())
    dest = _expand_path(dest, project_name(), exists=False)
    return hdfs.move(src, dest)


def rename(src, dest):
    """
    Rename src to dest.

    Args:
        :src: You can specify either a full hdfs pathname or a relative one (relative to your Project's path in HDFS).
        :dest: You can specify either a full hdfs pathname or a relative one (relative to your Project's path in HDFS).
    """
    src = _expand_path(src, project_name())
    dest = _expand_path(dest, project_name(), exists=False)
    return hdfs.rename(src, dest)


def chown(hdfs_path, user, group, project=None):
    """
    Change file owner and group.

    Args:
        :hdfs_path: You can specify either a full hdfs pathname or a relative one (relative to the given project path in HDFS).
        :user: New hdfs username
        :group: New hdfs group
        :project: If this value is not specified, it will get the path to your project. If you need to path to another project, you can specify the name of the project as a string.
    """
    if project == None:
        project = project_name()
    hdfs_path = _expand_path(hdfs_path, project)
    return hdfs.chown(hdfs_path, user, group)


def chmod(hdfs_path, mode, project=None):
    """
    Change file mode bits.

    Args:
        :hdfs_path: You can specify either a full hdfs pathname or a relative one (relative to your Project's path in HDFS).
        :mode: File mode (user/group/world privilege) bits
        :project: If this value is not specified, it will get the path to your project. If you need to path to another project, you can specify the name of the project as a string.
    """
    if project == None:
        project = project_name()
    hdfs_path = _expand_path(hdfs_path, project)
    return hdfs.chmod(hdfs_path, mode)


def stat(hdfs_path, project=None):
    """
    Performs the equivalent of os.stat() on path, returning a StatResult object.

    Args:
        :hdfs_path: If this value is not specified, it will get the path to your project. You can specify either a full hdfs pathname or a relative one (relative to your Project's path in HDFS).
        :project: If this value is not specified, it will get the path to your project. If you need to path to another project, you can specify the name of the project as a string.

    Returns:
        StatResult object
    """
    if project == None:
        project = project_name()
    hdfs_path = _expand_path(hdfs_path, project)
    return hdfs.stat(hdfs_path)


def access(hdfs_path, mode, project=None):
    """
    Perform the equivalent of os.access() on path.

    Args:
        :hdfs_path: You can specify either a full hdfs pathname or a relative one (relative to your Project's path in HDFS).
        :mode: File mode (user/group/world privilege) bits
        :project: If this value is not specified, it will get the path to your project. If you need to path to another project, you can specify the name of the project as a string.

    Returns:
        True if access is allowed, False if not.
    """
    if project == None:
        project = project_name()
    hdfs_path = _expand_path(hdfs_path, project)
    return hdfs.access(hdfs_path, mode)


def _mkdir_p(path):
    """
    Creates path on local filesystem

    Args:
        path to create

    Raises:
        OSError
    """
    try:
        os.makedirs(path)
    except OSError as exc:
        if exc.errno == errno.EEXIST and os.path.isdir(path):
            pass
        else:
            raise


def open_file(hdfs_path, project=None, flags='rw', buff_size=0):
    """
    Opens an HDFS file for read/write/append and returns a file descriptor object (fd) that should be closed when no longer needed.

    Args:
        hdfs_path: you can specify either a full hdfs pathname or a relative one (relative to your project's path in HDFS)
        flags: supported opening modes are 'r', 'w', 'a'. In addition, a trailing 't' can be added to specify text mode (e.g, 'rt' = open for reading text)
        buff_size: Pass 0 as buff_size if you want to use the "configured" values, i.e the ones set in the Hadoop configuration files.

    Returns:
        A file descriptor (fd) that needs to be closed (fd-close()) when it is no longer needed.

    Raises:
        IOError: If the file does not exist.
    """
    if project == None:
        project = project_name()
    hdfs_path = _expand_path(hdfs_path, project, exists=False)
    fs_handle = get_fs()
    fd = fs_handle.open_file(hdfs_path, flags, buff_size=buff_size)
    return fd


def close():
    """
    Closes an the HDFS connection (disconnects to the namenode)
    """
    hdfs.close()


def exists(hdfs_path, project=None):
    """
    Return True if hdfs_path exists in the default HDFS.

    Args:
        :hdfs_path: You can specify either a full hdfs pathname or a relative one (relative to your Project's path in HDFS).
        :project: If this value is not specified, it will get the path to your project. If you need to path to another project, you can specify the name of the project as a string.


    Returns:
        True if hdfs_path exists.

    Raises: IOError
    """
    if project == None:
        project = project_name()

    try:
        hdfs_path = _expand_path(hdfs_path, project)
    except IOError:
        return False
    return hdfs.path.exists(hdfs_path)


def isdir(hdfs_path, project=None):
    """
    Return True if path refers to a directory.

    Args:
        :hdfs_path: You can specify either a full hdfs pathname or a relative one (relative to your Project's path in HDFS).
        :project: If this value is not specified, it will get the path to your project. If you need to path to another project, you can specify the name of the project as a string.

    Returns:
        True if path refers to a directory.

    Raises: IOError
    """
    if project == None:
        project = project_name()
    hdfs_path = _expand_path(hdfs_path, project)
    return hdfs.isdir(hdfs_path)


def isfile(hdfs_path, project=None):
    """
    Return True if path refers to a file.

    Args:
        :hdfs_path: You can specify either a full hdfs pathname or a relative one (relative to your Project's path in HDFS).
        :project: If this value is not specified, it will get the path to your project. If you need to path to another project, you can specify the name of the project as a string.

    Returns:
        True if path refers to a file.

    Raises: IOError
    """
    if project == None:
        project = project_name()
    hdfs_path = _expand_path(hdfs_path, project)
    return path.isfile(hdfs_path)

def isdir(hdfs_path, project=None):
    """
    Return True if path refers to a directory.

    Args:
        :hdfs_path: You can specify either a full hdfs pathname or a relative one (relative to your Project's path in HDFS).
        :project: If this value is not specified, it will get the path to your project. If you need to path to another project, you can specify the name of the project as a string.

    Returns:
        True if path refers to a file.

    Raises: IOError
    """
    if project == None:
        project = project_name()
    hdfs_path = _expand_path(hdfs_path, project)
    return path.isdir(hdfs_path)


def capacity():
    """
    Returns the raw capacity of the filesystem

    Returns:
        filesystem capacity (int)
    """
    return hdfs.capacity()


def dump(data, hdfs_path):
    """
    Dumps data to a file

    Args:
        :data: data to write to hdfs_path
        :hdfs_path: You can specify either a full hdfs pathname or a relative one (relative to your Project's path in HDFS).
    """

    hdfs_path = _expand_path(hdfs_path, exists=False)
    return hdfs.dump(data, hdfs_path)


def load(hdfs_path):
    """
    Read the content of hdfs_path and return it.

    Args:
        :hdfs_path: You can specify either a full hdfs pathname or a relative one (relative to your Project's path in HDFS).

    Returns:
        the read contents of hdfs_path
    """
    hdfs_path = _expand_path(hdfs_path)
    return hdfs.load(hdfs_path)

def ls(hdfs_path, recursive=False, exclude_nn_addr=False):
    """
    lists a directory in HDFS

    Args:
        :hdfs_path: You can specify either a full hdfs pathname or a relative one (relative to your Project's path in HDFS).

    Returns:
        returns a list of hdfs paths
    """
    if exclude_nn_addr:
        hdfs_path = re.sub(r"\d+.\d+.\d+.\d+:\d+", "", hdfs_path)
    hdfs_path = _expand_path(hdfs_path)
    return hdfs.ls(hdfs_path, recursive=recursive)

def stat(hdfs_path):
    """
    Performs the equivalent of os.stat() on hdfs_path, returning a StatResult object.

    Args:
        :hdfs_path: You can specify either a full hdfs pathname or a relative one (relative to your Project's path in HDFS).

    Returns:
        returns a list of hdfs paths
    """
    hdfs_path = _expand_path(hdfs_path)
    return hdfs.stat(hdfs_path)

def abs_path(hdfs_path):
    """
     Return an absolute path for hdfs_path.

     Args:
         :hdfs_path: You can specify either a full hdfs pathname or a relative one (relative to your Project's path in HDFS).

    Returns:
        Return an absolute path for hdfs_path.
    """
    return _expand_path(hdfs_path)

def add_module(hdfs_path, project=None):
    """
     Add a .py or .ipynb file from HDFS to sys.path

     For example, if you execute:

     >>> add_module("Resources/my_module.py")
     >>> add_module("Resources/my_notebook.ipynb")

     You can import it simply as:

     >>> import my_module
     >>> import my_notebook

     Args:
         :hdfs_path: You can specify either a full hdfs pathname or a relative one (relative to your Project's path in HDFS) to a .py or .ipynb file

     Returns:
        Return full local path to localized python file or converted python file in case of .ipynb file
    """

    localized_deps = os.getcwd() + "/localized_deps"
    if not os.path.exists(localized_deps):
        os.mkdir(localized_deps)
        open(localized_deps + '/__init__.py', mode='w').close()

    if localized_deps not in sys.path:
        sys.path.append(localized_deps)

    if project == None:
        project = project_name()
    hdfs_path = _expand_path(hdfs_path, project)

    if path.isfile(hdfs_path) and hdfs_path.endswith('.py'):
        py_path = copy_to_local(hdfs_path, localized_deps, overwrite=True)
        if py_path not in sys.path:
            sys.path.append(py_path)
        _reload(py_path)
        return py_path
    elif path.isfile(hdfs_path) and hdfs_path.endswith('.ipynb'):
        ipynb_path = copy_to_local(hdfs_path, localized_deps, overwrite=True)
        python_path = sys.executable
        jupyter_binary = os.path.dirname(python_path) + '/jupyter'
        if not os.path.exists(jupyter_binary):
            raise Exception('Could not find jupyter binary on path {}'.format(jupyter_binary))

        converted_py_path = os.path.splitext(ipynb_path)[0] + '.py'
        if os.path.exists(converted_py_path):
            os.remove(converted_py_path)

        conversion = subprocess.Popen([jupyter_binary, 'nbconvert', '--to', 'python', ipynb_path], stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        out, err = conversion.communicate()
        if conversion.returncode != 0:
            raise Exception("Notebook conversion to .py failed: stdout: {} \n stderr: {}".format(out, err))

        if not os.path.exists(converted_py_path):
            raise Exception('Could not find converted .py file on path {}'.format(converted_py_path))
        if converted_py_path not in sys.path:
            sys.path.append(converted_py_path)
        _reload(converted_py_path)
        return converted_py_path
    else:
        raise Exception("Given path " + hdfs_path + " does not point to a .py or .ipynb file")

def _reload(path):
    try:
        module_name = ntpath.basename(path).split(".")[0]
        imported_module = importlib.import_module(module_name)
        importlib.reload(imported_module)
    except Exception as err:
        log.error('Failed to automatically reload module on path {} with exception: {}'.format(path, err))

def is_tls_enabled():
    """
    Reads the ipc.server.ssl.enabled property from core-site.xml.

    Returns:
        returns True if ipc.server.ssl.enabled is true. False otherwise.
    """
    global tls_enabled
    if tls_enabled is None:
        hadoop_conf_path = os.environ['HADOOP_CONF_DIR']
        xmldoc = minidom.parse(os.path.join(hadoop_conf_path,'core-site.xml'))
        itemlist = xmldoc.getElementsByTagName('property')
        for item in itemlist:
            name = item.getElementsByTagName("name")[0]
            if name.firstChild.data == "ipc.server.ssl.enabled":
                tls_enabled = item.getElementsByTagName("value")[0].firstChild.data == 'true'
    return tls_enabled

def _get_webhdfs_address():
    """
    Makes an SRV DNS query to get the target and port of NameNode's web interface

    Returns:
        returns webhdfs endpoint
    """
    global webhdfs_address
    if webhdfs_address is None:
        _, port = ServiceDiscovery.get_any_service('http.namenode')
        webhdfs_address = ServiceDiscovery.construct_service_fqdn('http.namenode') + ":" + str(port)
    return webhdfs_address
        
def get_webhdfs_host():
    """
    Makes an SRV DNS query and gets the actual hostname of the NameNode

    Returns:
        returns NameNode's hostname
    """
    return ServiceDiscovery.construct_service_fqdn('http.namenode')

def get_webhdfs_port():
    """
    Makes an SRV DNS query and gets NameNode's port for WebHDFS

    Returns:
        returns NameNode's port for WebHDFS
    """
    return _get_webhdfs_address().split(":")[1]



"""
API for attaching, detaching, and reading extended metadata to HopsFS files/directories.
It uses the Hopsworks /xattrs REST API
"""
from hops import constants, util, hdfs
from hops.exceptions import RestAPIError
import urllib


def set_xattr(hdfs_path, xattr_name, value):
    """
    Attach an extended attribute to an hdfs_path

    Args:
        :hdfs_path: path of a file or directory
        :xattr_name: name of the extended attribute
        :value: value of the extended attribute

    Returns:
        None
    """
    value = str(value)
    hdfs_path = urllib.parse.quote(hdfs._expand_path(hdfs_path))
    headers = {constants.HTTP_CONFIG.HTTP_CONTENT_TYPE: constants.HTTP_CONFIG.HTTP_APPLICATION_JSON}
    method = constants.HTTP_CONFIG.HTTP_PUT
    resource_url = constants.DELIMITERS.SLASH_DELIMITER + \
                   constants.REST_CONFIG.HOPSWORKS_REST_RESOURCE + constants.DELIMITERS.SLASH_DELIMITER + \
                   constants.REST_CONFIG.HOPSWORKS_PROJECT_RESOURCE + constants.DELIMITERS.SLASH_DELIMITER + \
                   hdfs.project_id() + constants.DELIMITERS.SLASH_DELIMITER + \
                   constants.REST_CONFIG.HOPSWORKS_XATTR_RESOURCE + constants.DELIMITERS.SLASH_DELIMITER + \
                   hdfs_path + constants.DELIMITERS.QUESTION_MARK_DELIMITER + constants.XATTRS.XATTRS_PARAM_NAME + \
                   constants.DELIMITERS.JDBC_CONNECTION_STRING_VALUE_DELIMITER + xattr_name
    response = util.send_request(method, resource_url, data=value, headers=headers)
    response_object = response.json()
    if response.status_code >= 400:
        error_code, error_msg, user_msg = util._parse_rest_error(response_object)
        raise RestAPIError("Could not attach extened attributes from a path (url: {}), server response: \n " \
                           "HTTP code: {}, HTTP reason: {}, error code: {}, error msg: {}, user msg: {}".format(
            resource_url, response.status_code, response.reason, error_code, error_msg, user_msg))


def get_xattr(hdfs_path, xattr_name=None):
    """
    Get the extended attribute attached to an hdfs_path.

    Args:
        :hdfs_path: path of a file or directory
        :xattr_name: name of the extended attribute

    Returns:
        A dictionary with the extended attribute(s) as key value pair(s). If the :xattr_name is None,
         the API returns all associated extended attributes.
    """
    hdfs_path = urllib.parse.quote(hdfs._expand_path(hdfs_path))
    headers = {constants.HTTP_CONFIG.HTTP_CONTENT_TYPE: constants.HTTP_CONFIG.HTTP_APPLICATION_JSON}
    method = constants.HTTP_CONFIG.HTTP_GET
    resource_url = constants.DELIMITERS.SLASH_DELIMITER + \
                   constants.REST_CONFIG.HOPSWORKS_REST_RESOURCE + constants.DELIMITERS.SLASH_DELIMITER + \
                   constants.REST_CONFIG.HOPSWORKS_PROJECT_RESOURCE + constants.DELIMITERS.SLASH_DELIMITER + \
                   hdfs.project_id() + constants.DELIMITERS.SLASH_DELIMITER + \
                   constants.REST_CONFIG.HOPSWORKS_XATTR_RESOURCE + constants.DELIMITERS.SLASH_DELIMITER + \
                   hdfs_path
    if xattr_name is not None:
        resource_url += constants.DELIMITERS.QUESTION_MARK_DELIMITER + constants.XATTRS.XATTRS_PARAM_NAME + \
                        constants.DELIMITERS.JDBC_CONNECTION_STRING_VALUE_DELIMITER + xattr_name

    response = util.send_request(method, resource_url, headers=headers)
    response_object = response.json()
    if response.status_code >= 400:
        error_code, error_msg, user_msg = util._parse_rest_error(response_object)
        raise RestAPIError("Could not get extened attributes attached to a path (url: {}), server response: \n " \
                           "HTTP code: {}, HTTP reason: {}, error code: {}, error msg: {}, user msg: {}".format(
            resource_url, response.status_code, response.reason, error_code, error_msg, user_msg))

    results = {}
    for item in response_object["items"]:
        results[item["name"]] = item["value"]
    return results


def remove_xattr(hdfs_path, xattr_name):
    """
    Remove an extended attribute attached to an hdfs_path

    Args:
        :hdfs_path: path of a file or directory
        :xattr_name: name of the extended attribute

    Returns:
        None
    """
    hdfs_path = urllib.parse.quote(hdfs._expand_path(hdfs_path))
    headers = {constants.HTTP_CONFIG.HTTP_CONTENT_TYPE: constants.HTTP_CONFIG.HTTP_APPLICATION_JSON}
    method = constants.HTTP_CONFIG.HTTP_DELETE
    resource_url = constants.DELIMITERS.SLASH_DELIMITER + \
                   constants.REST_CONFIG.HOPSWORKS_REST_RESOURCE + constants.DELIMITERS.SLASH_DELIMITER + \
                   constants.REST_CONFIG.HOPSWORKS_PROJECT_RESOURCE + constants.DELIMITERS.SLASH_DELIMITER + \
                   hdfs.project_id() + constants.DELIMITERS.SLASH_DELIMITER + \
                   constants.REST_CONFIG.HOPSWORKS_XATTR_RESOURCE + constants.DELIMITERS.SLASH_DELIMITER + \
                   hdfs_path + constants.DELIMITERS.QUESTION_MARK_DELIMITER + constants.XATTRS.XATTRS_PARAM_NAME + \
                   constants.DELIMITERS.JDBC_CONNECTION_STRING_VALUE_DELIMITER + xattr_name
    response = util.send_request(method, resource_url, headers=headers)
    if response.status_code >= 400:
        response_object = response.json()
        error_code, error_msg, user_msg = util._parse_rest_error(response_object)
        raise RestAPIError("Could not remove extened attributes from a path (url: {}), server response: \n " \
                           "HTTP code: {}, HTTP reason: {}, error code: {}, error msg: {}, user msg: {}".format(
            resource_url, response.status_code, response.reason, error_code, error_msg, user_msg))


from pyhive import hive
import os
from . import tls

def setup_hive_connection(database=None):
    """
    export enviroment variable with Hive connection configuration
    so it can be used by ipython-sql and PyHive to establish a connection with Hive

    Args:
        :database: Name of the database to connect to

    """

    if not ('HIVE_ENDPOINT' in os.environ and 'PROJECT_NAME' in os.environ) : 
        raise EnvironmentError("HIVE_ENDPOINT or PROJECT_NAME is not exported." + 
            "The Hive module is meant to be run only in the context of a Python kernel") 

    connection_conf = {
        'auth' : 'CERTIFICATES',
        'keystore' : tls.get_key_store(),
        'truststore' : tls.get_trust_store(),
        'keystore_password' : tls._get_cert_pw()
    }

    database_name = database if database else os.environ['PROJECT_NAME']

    os.environ['DATABASE_URL'] = "hive://" + os.environ['HIVE_ENDPOINT'] + "/" + database_name + '?'
    os.environ['DATABASE_URL'] = os.environ['DATABASE_URL'] + '&'.join(['%s=%s' % (key, value) for (key, value) in connection_conf.items()])



"""

Miscellaneous utility functions for user applications.

"""

import os
import signal
import urllib
from ctypes import cdll
import socket
import warnings

import json

import boto3

from hops import constants
from hops import tls

from cryptography import x509
from cryptography.x509.oid import NameOID
from hops.exceptions import UnkownSecretStorageError, APIKeyFileNotFound
import base64
from socket import socket
from json.decoder import JSONDecodeError
from hops.exceptions import RestAPIError
from hops.exceptions import APIKeyFileNotFound


verify = None
#! Needed for hops library backwards compatability
try:
    import requests
    import urllib3
    urllib3.disable_warnings(urllib3.exceptions.SecurityWarning)
    urllib3.disable_warnings(urllib3.exceptions.InsecureRequestWarning)
except:
    pass

# in case importing in %%local
try:
    from pyspark.sql import SparkSession
except:
    pass

session = requests.session()


def http(resource_url, headers=None, method=constants.HTTP_CONFIG.HTTP_GET, data=None):
    response = send_request(method, resource_url, headers=headers, data=data)
    try:
        response_object = response.json()
    except JSONDecodeError:
        response_object = None

    if (response.status_code // 100) != 2:
        if response_object:
            error_code, error_msg, user_msg = _parse_rest_error(response_object)
        else:
            error_code, error_msg, user_msg = "", "", ""

        raise RestAPIError("Could not execute HTTP request (url: {}), server response: \n "
                           "HTTP code: {}, HTTP reason: {}, error code: {}, error msg: {}, user msg: {}".format(
            resource_url, response.status_code, response.reason, error_code, error_msg, user_msg))

    return response_object


def _get_elastic_endpoint():
    """

    Returns:
        The endpoint for putting things into elastic search

    """
    elastic_endpoint = os.environ[constants.ENV_VARIABLES.ELASTIC_ENDPOINT_ENV_VAR]
    host, port = elastic_endpoint.split(':')
    return host + ':' + port

def _get_hopsworks_rest_endpoint():
    """

    Returns:
        The hopsworks REST endpoint for making requests to the REST API

    """
    return os.environ[constants.ENV_VARIABLES.REST_ENDPOINT_END_VAR]

def _get_host_port_pair():
    """
    Removes "http or https" from the rest endpoint and returns a list
    [endpoint, port], where endpoint is on the format /path.. without http://

    Returns:
        a list [endpoint, port]
    """
    endpoint = _get_hopsworks_rest_endpoint()
    if 'http' in endpoint:
        last_index = endpoint.rfind('/')
        endpoint = endpoint[last_index + 1:]
    host_port_pair = endpoint.split(':')
    return host_port_pair


def connect(host=None, port=443, scheme="https", hostname_verification=False,
            api_key_file=None,
            region_name=constants.AWS.DEFAULT_REGION,
            secrets_store=constants.LOCAL.LOCAL_STORE,
            trust_store_path=None):
    """
    Connect to a Hopworks instance. Sets the REST API endpoint and any necessary authentication parameters.

    Example usage:

    >>> from hops import util
    >>> util.connect("localhost", api_key_file="api_key_file")

    Args:
        :host: the hostname of the Hopsworks cluster. If none specified, the library will attempt to the one set by the environment variable constants.ENV_VARIABLES.REST_ENDPOINT_END_VAR
        :port: the REST port of the Hopsworks cluster
        :scheme: the scheme to use for connection to the REST API.
        :hostname_verification: whether or not to verify Hopsworks' certificate - default True
        :api_key: path to a file containing an API key or the actual API key value. For secrets_store=local only.
        :region_name: The name of the AWS region in which the required secrets are stored
        :secrets_store: The secrets storage to be used. Secretsmanager or parameterstore for AWS, local otherwise.
        :trust_store_path: path to the file  containing the Hopsworks certificates

    Returns:
        None
    """
    if host is not None:
        os.environ[constants.ENV_VARIABLES.REST_ENDPOINT_END_VAR] = scheme + "://" + host + ":" + str(port)

    os.environ[constants.ENV_VARIABLES.REGION_NAME_ENV_VAR] = region_name
    if secrets_store == constants.LOCAL.LOCAL_STORE and not api_key_file:
        warnings.warn("API key was not provided and secrets_store is local. "
                      "When the connect method is used outside of a Hopsworks instance, it is recommended to "
                      "use an API key. Falling back to JWT...")
    else:
        try:
            os.environ[constants.ENV_VARIABLES.API_KEY_ENV_VAR] = get_secret(secrets_store, 'api-key', api_key_file)
            if constants.ENV_VARIABLES.SERVING_API_KEY_ENV_VAR not in os.environ:
                os.environ[constants.ENV_VARIABLES.SERVING_API_KEY_ENV_VAR] = os.environ[constants.ENV_VARIABLES.API_KEY_ENV_VAR]
        except APIKeyFileNotFound:
            warnings.warn("API key file was not found. Will use the provided api_key value as the API key")
            os.environ[constants.ENV_VARIABLES.API_KEY_ENV_VAR] = api_key_file

    if trust_store_path is not None:
        os.environ[constants.ENV_VARIABLES.DOMAIN_CA_TRUSTSTORE_PEM_ENV_VAR] = trust_store_path

    os.environ[constants.ENV_VARIABLES.REQUESTS_VERIFY_ENV_VAR] = str(hostname_verification).lower()


def set_auth_header(headers):
    """
    Set authorization header for HTTP requests to Hopsworks, depending if setup is remote or not.

    Args:
        http headers
    """
    if constants.HTTP_CONFIG.HTTP_AUTHORIZATION in headers:
        return  # auth header already set

    if constants.ENV_VARIABLES.API_KEY_ENV_VAR in os.environ:
        headers[constants.HTTP_CONFIG.HTTP_AUTHORIZATION] = "ApiKey " + \
            os.environ[constants.ENV_VARIABLES.API_KEY_ENV_VAR]
    else:
        headers[constants.HTTP_CONFIG.HTTP_AUTHORIZATION] = "Bearer " + get_jwt()


def get_requests_verify(hostname, port):
    """
    Returns:
        if env var HOPS_UTIL_VERIFY is not false
            if the env variable is set, then use the certificate, otherwise return true
        return false
    """
    if constants.ENV_VARIABLES.REQUESTS_VERIFY_ENV_VAR in os.environ and os.environ[
        constants.ENV_VARIABLES.REQUESTS_VERIFY_ENV_VAR] == 'true':

        try:
            if constants.ENV_VARIABLES.DOMAIN_CA_TRUSTSTORE_ENV_VAR in os.environ:
                # need to convert the jks to pem
                return tls.get_ca_chain_location()
            elif constants.ENV_VARIABLES.DOMAIN_CA_TRUSTSTORE_PEM_ENV_VAR in os.environ: 
                return os.environ[constants.ENV_VARIABLES.DOMAIN_CA_TRUSTSTORE_PEM_ENV_VAR]
            else:
                return True
        except x509.ExtensionNotFound:
            return True

    return False


def send_request(method, resource, data=None, headers=None, stream=False, files=None):
    """
    Sends a request to Hopsworks. In case of Unauthorized response, submit the request once more as jwt might not
    have been read properly from local container.
    Args:
        method: HTTP(S) method
        resource: Hopsworks resource
        data: HTTP(S) payload
        headers: HTTP(S) headers
        stream: set the stream for the session object
        files: dictionary of {filename: fileobject} files to multipart upload.
    Returns:
        HTTP(S) response
    """
    if headers is None:
        headers = {}
    global verify
    host, port = _get_host_port_pair()
    if verify is None:
        verify = get_requests_verify(host, port)
    set_auth_header(headers)
    url = _get_hopsworks_rest_endpoint() + resource
    req = requests.Request(method, url, data=data, headers=headers, files=files)
    prepped = session.prepare_request(req)

    response = session.send(prepped, verify=verify, stream=stream)

    if response.status_code == constants.HTTP_CONFIG.HTTP_UNAUTHORIZED:
        set_auth_header(headers)
        prepped = session.prepare_request(req)
        response = session.send(prepped, stream=stream)
    return response

def _parse_rest_error(response_dict):
    """
    Parses a JSON response from hopsworks after an unsuccessful request

    Args:
        response_dict: the JSON response represented as a dict

    Returns:
        error_code, error_msg, user_msg
    """
    error_code = -1
    error_msg = ""
    user_msg = ""
    if constants.REST_CONFIG.JSON_ERROR_CODE in response_dict:
        error_code = response_dict[constants.REST_CONFIG.JSON_ERROR_CODE]
    if constants.REST_CONFIG.JSON_ERROR_MSG in response_dict:
        error_msg = response_dict[constants.REST_CONFIG.JSON_ERROR_MSG]
    if constants.REST_CONFIG.JSON_USR_MSG in response_dict:
        user_msg = response_dict[constants.REST_CONFIG.JSON_USR_MSG]
    return error_code, error_msg, user_msg

def get_job_name():
    """
    If this method is called from inside a hopsworks job, it returns the name of the job.

    Returns:
        the name of the hopsworks job

    """
    if constants.ENV_VARIABLES.JOB_NAME_ENV_VAR in os.environ:
        return os.environ[constants.ENV_VARIABLES.JOB_NAME_ENV_VAR]

def get_jwt():
    """
    Retrieves jwt from local container.

    Returns:
        Content of jwt.token file in local container.
    """
    jwt_path = ""
    if constants.ENV_VARIABLES.SECRETS_DIR_ENV_VAR in os.environ:
        jwt_path = os.environ[constants.ENV_VARIABLES.SECRETS_DIR_ENV_VAR]

    jwt_path = os.path.join(jwt_path, constants.REST_CONFIG.JWT_TOKEN)

    with open(jwt_path, "r") as jwt:
        return jwt.read()

def parse_redhift_jdbc_url(url):
    """
    Parses a Redshift JDBC URL and extracts region_name, cluster_identifier, database and user.

    Args:
        :url: the JDBC URL

    Returns:
        region_name, cluster_identifier, database, user
    """

    jdbc_url = urllib.parse.urlparse(url)
    redshift_url = urllib.parse.urlparse(jdbc_url.path)
    if redshift_url.scheme != 'redshift':
        raise Exception('Trying to parse non-redshift url: ' + jdbc_url)
    cluster_identifier = redshift_url.netloc.split('.')[0]
    region_name = redshift_url.netloc.split('.')[2]
    database = redshift_url.path.split('/')[1]
    user = urllib.parse.parse_qs(jdbc_url.query)['user'][0]
    return region_name, cluster_identifier, database, user


def get_redshift_username_password(region_name, cluster_identifier, user, database):
    """
    Requests temporary Redshift credentials with a validity of 3600 seconds and the given parameters.

    Args:
        :region_name: the AWS region name
        :cluster_identifier: the Redshift cluster identifier
        :user: the Redshift user to get credentials for
        :database: the Redshift database

    Returns:
        user, password
    """

    client = boto3.client('redshift', region_name=region_name)
    credential = client.get_cluster_credentials(
        DbUser=user,
        DbName=database,
        ClusterIdentifier=cluster_identifier,
        DurationSeconds=3600,
        AutoCreate=False
    )
    return credential['DbUser'], credential['DbPassword']


def get_flink_conf_dir():
    """
    Returns the Flink configuration directory.

    Returns:
        The Flink config dir path.
    """
    if constants.ENV_VARIABLES.FLINK_CONF_DIR in os.environ:
        return os.environ[constants.ENV_VARIABLES.FLINK_CONF_DIR]


def get_flink_lib_dir():
    """
    Returns the Flink libraries directory.

    Returns:
        The Flink libraries dir path.
    """
    if constants.ENV_VARIABLES.FLINK_LIB_DIR in os.environ:
        return os.environ[constants.ENV_VARIABLES.FLINK_LIB_DIR]


def get_hadoop_home():
    """
    Returns the Hadoop home directory.

    Returns:
        The Hadoop home dir path.
    """
    if constants.ENV_VARIABLES.HADOOP_HOME in os.environ:
        return os.environ[constants.ENV_VARIABLES.HADOOP_HOME]


def get_hadoop_classpath_glob():
    """
    Returns the Hadoop glob classpath.

    Returns:
        The Hadoop glob classpath.
    """
    if constants.ENV_VARIABLES.HADOOP_CLASSPATH_GLOB in os.environ:
        return os.environ[constants.ENV_VARIABLES.HADOOP_CLASSPATH_GLOB]


def _validate_enable_online_featuregroup_schema(featuregroup_schema):
    """
    Validates the user-provided schema of an online feature group
    Args:
        :featuregroup_schema: the schema dict to validate

    Returns:
        schema with default values
    """
    if featuregroup_schema == None or len(featuregroup_schema) == 0:
        raise ValueError("The feature schema is invalid, featuregroup schema is empty: {} ".format(featuregroup_schema))
    primary_idx = -1
    for idx, feature_def in enumerate(featuregroup_schema):
        if constants.REST_CONFIG.JSON_FEATURE_DESCRIPTION not in feature_def or \
                        feature_def[constants.REST_CONFIG.JSON_FEATURE_DESCRIPTION] is None:
            feature_def[constants.REST_CONFIG.JSON_FEATURE_DESCRIPTION] = "-"
        if constants.REST_CONFIG.JSON_FEATURE_PARTITION not in feature_def or \
                feature_def[constants.REST_CONFIG.JSON_FEATURE_PARTITION] is None:
            feature_def[constants.REST_CONFIG.JSON_FEATURE_PARTITION] = False
        if constants.REST_CONFIG.JSON_FEATURE_PRIMARY in feature_def and \
                        feature_def[constants.REST_CONFIG.JSON_FEATURE_PRIMARY] is not None:
            primary_idx = idx
        if constants.REST_CONFIG.JSON_FEATURE_ONLINE_TYPE not in feature_def or \
                        feature_def[constants.REST_CONFIG.JSON_FEATURE_ONLINE_TYPE] is None:
            if constants.REST_CONFIG.JSON_FEATURE_TYPE not in feature_def or \
                            feature_def[constants.REST_CONFIG.JSON_FEATURE_TYPE] is None:
                feature_def[constants.REST_CONFIG.JSON_FEATURE_ONLINE_TYPE] = \
                    feature_def[constants.REST_CONFIG.JSON_FEATURE_TYPE]
            else:
                raise ValueError("The feature schema is invalid, the feature definition: {} "
                                 "does not contain a type".format(feature_def))

        if constants.REST_CONFIG.JSON_FEATURE_NAME not in feature_def or \
                        feature_def[constants.REST_CONFIG.JSON_FEATURE_NAME] is None:
            raise ValueError("The feature schema is invalid, the feature definition: {} "
                             "does not contain a name".format(feature_def))
    if primary_idx == -1:
        raise ValueError("You must mark at least one feature as primary in the online feature group")

    return featuregroup_schema

def num_executors():
    """
    Get the number of executors configured for Jupyter

    Returns:
        Number of configured executors for Jupyter
    """
    sc = _find_spark().sparkContext
    try:
        return int(sc._conf.get("spark.dynamicAllocation.maxExecutors"))
    except:
        raise RuntimeError('Failed to find spark.dynamicAllocation.maxExecutors property, please select your mode as either Experiment, Parallel Experiments or Distributed Training.')

def num_param_servers():
    """
    Get the number of parameter servers configured for Jupyter

    Returns:
        Number of configured parameter servers for Jupyter
    """
    sc = _find_spark().sparkContext
    try:
        return int(os.environ['NUM_TF_PS'])
    except:
        return 0

def _find_spark():
    """
    Returns: SparkSession
    """
    return SparkSession.builder.getOrCreate()

def _on_executor_exit(signame):
    """
    Return a function to be run in a child process which will trigger
    SIGNAME to be sent when the parent process dies

    Args:
        :signame: the signame to send

    Returns:
        set_parent_exit_signal
    """
    signum = getattr(signal, signame)
    def set_parent_exit_signal():
        # http://linux.die.net/man/2/prctl

        PR_SET_PDEATHSIG = 1
        result = cdll['libc.so.6'].prctl(PR_SET_PDEATHSIG, signum)
        if result != 0:
            raise Exception('prctl failed with error code %s' % result)
    return set_parent_exit_signal


def _assumed_role():
    client = boto3.client('sts')
    response = client.get_caller_identity()
    # arns for assumed roles in SageMaker follow the following schema
    # arn:aws:sts::123456789012:assumed-role/my-role-name/my-role-session-name
    local_identifier = response['Arn'].split(':')[-1].split('/')
    if len(local_identifier) != 3 or local_identifier[0] != 'assumed-role':
        raise Exception(
            'Failed to extract assumed role from arn: ' + response['Arn'])
    return local_identifier[1]

def _get_region():
    if (os.environ[constants.ENV_VARIABLES.REGION_NAME_ENV_VAR] != constants.AWS.DEFAULT_REGION):
        return os.environ[constants.ENV_VARIABLES.REGION_NAME_ENV_VAR]
    else:
        return None

def _query_secrets_manager(secret_key):
    secret_name = 'hopsworks/role/' + _assumed_role()
    args = {'service_name': 'secretsmanager'}
    region_name = _get_region()
    if region_name:
        args['region_name'] = region_name
    client = boto3.client(**args)
    get_secret_value_response = client.get_secret_value(SecretId=secret_name)
    return json.loads(get_secret_value_response['SecretString'])[secret_key]

def _query_parameter_store(secret_key):
    args = {'service_name': 'ssm'}
    region_name = _get_region()
    if region_name:
        args['region_name'] = region_name
    client = boto3.client(**args)
    name = '/hopsworks/role/' + _assumed_role() + '/type/' + secret_key
    return client.get_parameter(Name=name, WithDecryption=True)['Parameter']['Value']

def get_secret(secrets_store, secret_key=None, api_key_file=None):
    """
    Returns secret value from the AWS Secrets Manager or Parameter Store

    Args:
        :secrets_store: the underlying secrets storage to be used, e.g. `secretsmanager` or `parameterstore`
        :secret_type (str): key for the secret value, e.g. `api-key`, `cert-key`, `trust-store`, `key-store`
        :api_token_file: path to a file containing an api key
    Returns:
        :str: secret value
    """
    if secrets_store == constants.AWS.SECRETS_MANAGER:
        return _query_secrets_manager(secret_key)
    elif secrets_store == constants.AWS.PARAMETER_STORE:
        return _query_parameter_store(secret_key)
    elif secrets_store == constants.LOCAL.LOCAL_STORE:
        if not api_key_file:
            raise Exception('api_key_file needs to be set for local mode')
        try:
            with open(api_key_file) as f:
                return f.readline().strip()
        except:
            raise APIKeyFileNotFound('API Key fiel could not be read or was not found')
    else:
        raise UnkownSecretStorageError(
            "Secrets storage " + secrets_store + " is not supported.")

def write_b64_cert_to_bytes(b64_string, path):
    """Converts b64 encoded certificate to bytes file .

    Args:
        :b64_string (str): b64 encoded string of certificate
        :path (str): path where file is saved, including file name. e.g. /path/key-store.jks
    """

    with open(path, 'wb') as f:
        cert_b64 = base64.b64decode(b64_string)
        f.write(cert_b64)

def attach_jupyter_configuration_to_notebook(kernel_id):
    method = constants.HTTP_CONFIG.HTTP_PUT
    resource_url = constants.DELIMITERS.SLASH_DELIMITER + \
                   constants.REST_CONFIG.HOPSWORKS_REST_RESOURCE + constants.DELIMITERS.SLASH_DELIMITER + \
                   "project" + constants.DELIMITERS.SLASH_DELIMITER + os.environ["HOPSWORKS_PROJECT_ID"] + \
                   constants.DELIMITERS.SLASH_DELIMITER + "jupyter/attachConfiguration" + \
                   constants.DELIMITERS.SLASH_DELIMITER + os.environ["HADOOP_USER_NAME"] + \
                   constants.DELIMITERS.SLASH_DELIMITER + kernel_id
    headers = {constants.HTTP_CONFIG.HTTP_CONTENT_TYPE: constants.HTTP_CONFIG.HTTP_APPLICATION_JSON}
    response = send_request(method, resource_url, headers=headers)
    if response.status_code >= 400:
        response_object = response.json()
        error_code, error_msg, user_msg = _parse_rest_error(response_object)
        raise Exception(error_msg)



from . import mirrored_reservation
import sys

if __name__ == "__main__":
  host = sys.argv[1]
  port = int(sys.argv[2])
  addr = (host, port)
  client = mirrored_reservation.Client(addr)
  client.request_stop()
  client.close()



"""
AWS temporary credential provider.

"""

from hops import constants, util, hdfs
from hops.exceptions import RestAPIError
import os


def assume_role(role_arn=None, role_session_name=None, duration_seconds=3600):
    """
    Assume a role and sets the temporary credential to the spark context hadoop configuration and environment variables.

    Args:
        :role_arn: (string) the role arn to be assumed
        :role_session_name: (string) use to uniquely identify a session when the same role is assumed by different principals or for different reasons.
        :duration_seconds: (int) the duration of the session. Maximum session duration is 3600 seconds.

    >>> from hops.credentials_provider import assume_role
    >>> assume_role(role_arn="arn:aws:iam::<AccountNumber>:role/analyst")

    or

    >>> assume_role() # to assume the default role

    Returns:
        temporary credentials
    """
    query = _query_string(role_arn, role_session_name, duration_seconds)
    method = constants.HTTP_CONFIG.HTTP_GET
    resource_url = _get_cloud_resource() + constants.REST_CONFIG.HOPSWORKS_AWS_CLOUD_SESSION_TOKEN_RESOURCE + query

    response = util.send_request(method, resource_url)
    json_content = _parse_response(response, resource_url)
    _set_spark_hadoop_conf(json_content)
    _set_envs(json_content)
    return json_content


def get_roles():
    """
    Get all roles mapped to the current project

    >>> from hops.credentials_provider import get_roles
    >>> get_roles()

    Returns:
        A list of role arn
    """
    json_content = _get_roles()
    items = json_content[constants.REST_CONFIG.JSON_ARRAY_ITEMS]
    roles = []
    for role in items:
        roles.append(role[constants.REST_CONFIG.JSON_CLOUD_ROLE])
    return roles


def get_role(role_id="default"):
    """
    Get a role arn mapped to the current project by id or if no id is supplied the default role is returned

    Args:
        :role_id: id of the role default

    >>> from hops.credentials_provider import get_role
    >>> get_role(id)
    or
    >>> get_role() # to get the default role

    Returns:
        A role arn
    """
    json_content = _get_roles(role_id=role_id)
    return json_content[constants.REST_CONFIG.JSON_CLOUD_ROLE]


def _set_spark_hadoop_conf(json_content):
    spark = None
    if constants.ENV_VARIABLES.SPARK_IS_DRIVER in os.environ:
        spark = util._find_spark()
    if spark is not None:
        sc = spark.sparkContext
        sc._jsc.hadoopConfiguration().set(constants.S3_CONFIG.S3_CREDENTIAL_PROVIDER_ENV,
                                          constants.S3_CONFIG.S3_TEMPORARY_CREDENTIAL_PROVIDER)
        sc._jsc.hadoopConfiguration().set(constants.S3_CONFIG.S3_ACCESS_KEY_ENV,
                                          json_content[constants.REST_CONFIG.JSON_ACCESS_KEY_ID])
        sc._jsc.hadoopConfiguration().set(constants.S3_CONFIG.S3_SECRET_KEY_ENV,
                                          json_content[constants.REST_CONFIG.JSON_SECRET_KEY_ID])
        sc._jsc.hadoopConfiguration().set(constants.S3_CONFIG.S3_SESSION_KEY_ENV,
                                          json_content[constants.REST_CONFIG.JSON_SESSION_TOKEN_ID])


def _set_envs(json_content):
    os.environ[constants.S3_CONFIG.AWS_ACCESS_KEY_ID_ENV] = json_content[constants.REST_CONFIG.JSON_ACCESS_KEY_ID]
    os.environ[constants.S3_CONFIG.AWS_SECRET_ACCESS_KEY_ENV] = json_content[constants.REST_CONFIG.JSON_SECRET_KEY_ID]
    os.environ[constants.S3_CONFIG.AWS_SESSION_TOKEN_ENV] = json_content[constants.REST_CONFIG.JSON_SESSION_TOKEN_ID]


def _get_roles(role_id=None):
    by_id = ""
    if role_id:
        by_id = constants.DELIMITERS.SLASH_DELIMITER + str(role_id)
    method = constants.HTTP_CONFIG.HTTP_GET
    resource_url = _get_cloud_resource() + constants.REST_CONFIG.HOPSWORKS_CLOUD_ROLE_MAPPINGS_RESOURCE + by_id
    response = util.send_request(method, resource_url)
    return _parse_response(response, resource_url)


def _parse_response(response, url):
    if response.ok:
        return response.json()
    else:
        error_code, error_msg, user_msg = util._parse_rest_error(response.json())
        raise RestAPIError("Error calling {}. Got status: HTTP code: {}, HTTP reason: {}, error code: {}, error msg: {}"
                           ", user msg: {}".format(url, response.status_code, response.reason, error_code, error_msg,
                                                   user_msg))


def _query_string(role, role_session_name, duration_seconds):
    query = ""
    if role:
        query = constants.REST_CONFIG.HOPSWORKS_CLOUD_SESSION_TOKEN_RESOURCE_QUERY_ROLE + "=" + role
    if role_session_name:
        if query != "":
            query += "&"
        query += constants.REST_CONFIG.HOPSWORKS_CLOUD_SESSION_TOKEN_RESOURCE_QUERY_SESSION + "=" + role_session_name
    if duration_seconds != 3600 and duration_seconds > 0:
        if query != "":
            query += "&"
        query += constants.REST_CONFIG.HOPSWORKS_CLOUD_SESSION_TOKEN_RESOURCE_QUERY_SESSION_DURATION + "=" + \
                 str(duration_seconds)

    if query != "":
        query = "?" + query
    return query


def _get_cloud_resource():
    return constants.DELIMITERS.SLASH_DELIMITER + constants.REST_CONFIG.HOPSWORKS_REST_RESOURCE + \
        constants.DELIMITERS.SLASH_DELIMITER + constants.REST_CONFIG.HOPSWORKS_PROJECT_RESOURCE + \
        constants.DELIMITERS.SLASH_DELIMITER + hdfs.project_id() + constants.DELIMITERS.SLASH_DELIMITER + \
        constants.REST_CONFIG.HOPSWORKS_CLOUD_RESOURCE + constants.DELIMITERS.SLASH_DELIMITER



"""
A module for setting up Elastcisearch spark connector.
"""

from hops import constants, util, hdfs, tls
from hops.exceptions import RestAPIError
import os


def _get_elasticsearch_url():
    return os.environ[constants.ENV_VARIABLES.ELASTIC_ENDPOINT_ENV_VAR]


def get_elasticsearch_index(index):
    """
    Get the valid elasticsearch index for later use. This helper method prefix the index name with the project name.

    Args:
        :index: the elasticsearch index to interact with.

    Returns:
        A valid elasticsearch index name.
    """
    return (hdfs.project_name() + "_" + index).lower()


def get_elasticsearch_config(index):
    """
    Get the required elasticsearch configuration to setup a connection using spark connector.

    Args:
        :index: the elasticsearch index to interact with.

    Returns:
        A dictionary with required configuration.
    """
    config = {
        constants.ELASTICSEARCH_CONFIG.SSL_CONFIG: "true",
        constants.ELASTICSEARCH_CONFIG.NODES: _get_elasticsearch_url(),
        constants.ELASTICSEARCH_CONFIG.NODES_WAN_ONLY: "true",
        constants.ELASTICSEARCH_CONFIG.SSL_KEYSTORE_LOCATION: tls.get_key_store(),
        constants.ELASTICSEARCH_CONFIG.SSL_KEYSTORE_PASSWORD: tls.get_key_store_pwd(),
        constants.ELASTICSEARCH_CONFIG.SSL_TRUSTSTORE_LOCATION: tls.get_trust_store(),
        constants.ELASTICSEARCH_CONFIG.SSL_TRUSTSTORE_PASSWORD: tls.get_trust_store_pwd(),
        constants.ELASTICSEARCH_CONFIG.HTTP_AUTHORIZATION: get_authorization_token(),
        constants.ELASTICSEARCH_CONFIG.INDEX: get_elasticsearch_index(index)
    }
    return config


def get_authorization_token():
    """
    Get the authorization token to interact with elasticsearch.

    Args:

    Returns:
        The authorization token to be used in http header.
    """
    headers = {constants.HTTP_CONFIG.HTTP_CONTENT_TYPE: constants.HTTP_CONFIG.HTTP_APPLICATION_JSON}
    method = constants.HTTP_CONFIG.HTTP_GET
    resource_url = constants.DELIMITERS.SLASH_DELIMITER + \
                   constants.REST_CONFIG.HOPSWORKS_REST_RESOURCE + constants.DELIMITERS.SLASH_DELIMITER + \
                   constants.REST_CONFIG.HOPSWORKS_ELASTIC_RESOURCE + constants.DELIMITERS.SLASH_DELIMITER + \
                   constants.REST_CONFIG.HOPSWORKS_ELASTIC_JWT_RESOURCE + constants.DELIMITERS.SLASH_DELIMITER + \
                   hdfs.project_id()
    response = util.send_request(method, resource_url, headers=headers)
    response_object = response.json()
    if response.status_code >= 400:
        error_code, error_msg, user_msg = util._parse_rest_error(response_object)
        raise RestAPIError("Could not get authorization token for elastic (url: {}), server response: \n " \
                           "HTTP code: {}, HTTP reason: {}, error code: {}, error msg: {}, user msg: {}".format(
            resource_url, response.status_code, response.reason, error_code, error_msg, user_msg))

    return "Bearer " + response_object["token"]



"""
Utility functions to retrieve information about available services and setting up security for the Hops platform.

These utils facilitates development by hiding complexity for programs interacting with Hops services.
"""

import os
from hops import devices, tensorboard, hdfs
from hops.experiment_impl.util import experiment_utils
from hops import util

import threading
import time
import socket
import json

from . import mirrored_reservation


def _run(sc, train_fn, run_id, local_logdir=False, name="no-name", evaluator=False):
    """

    Args:
        sc:
        train_fn:
        local_logdir:
        name:

    Returns:

    """
    app_id = str(sc.applicationId)

    num_executions = util.num_executors()

    #Each TF task should be run on 1 executor
    nodeRDD = sc.parallelize(range(num_executions), num_executions)

    #Make SparkUI intuitive by grouping jobs
    sc.setJobGroup(os.environ['ML_ID'], "{} | MirroredStrategy - Distributed Training".format(name))

    server = mirrored_reservation.Server(num_executions)
    server_addr = server.start()

    #Force execution on executor, since GPU is located on executor
    nodeRDD.foreachPartition(_prepare_func(app_id, run_id, train_fn, local_logdir, server_addr, evaluator, util.num_executors()))

    logdir = experiment_utils._get_logdir(app_id, run_id)

    print('Finished Experiment \n')

    path_to_return = logdir + '/.outputs.json'
    if hdfs.exists(path_to_return):
        with hdfs.open_file(path_to_return, flags="r") as fi:
            contents = fi.read()
            fi.close()
            return logdir, json.loads(contents)

    return logdir, None

def _prepare_func(app_id, run_id, train_fn, local_logdir, server_addr, evaluator, num_executors):
    """

    Args:
        app_id:
        run_id:
        train_fn:
        local_logdir:
        server_addr:

    Returns:

    """
    def _wrapper_fun(iter):
        """

        Args:
            iter:

        Returns:

        """

        for i in iter:
            executor_num = i

        experiment_utils._set_ml_id(app_id, run_id)

        t = threading.Thread(target=devices._print_periodic_gpu_utilization)
        if devices.get_num_gpus() > 0:
            t.start()

        is_chief = False
        logdir = None
        tb_hdfs_path = None
        try:
            host = experiment_utils._get_ip_address()

            tmp_socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            tmp_socket.bind(('', 0))
            port = tmp_socket.getsockname()[1]

            client = mirrored_reservation.Client(server_addr)
            host_port = host + ":" + str(port)

            client.register({"worker": host_port, "index": executor_num})
            cluster = client.await_reservations()
            tmp_socket.close()
            client.close()

            task_index = experiment_utils._find_index(host_port, cluster)

            if task_index == -1:
                cluster["task"] = {"type": "chief", "index": 0}
            else:
                cluster["task"] = {"type": "worker", "index": task_index}

            evaluator_node = None
            if evaluator:
                last_worker_index = len(cluster["cluster"]["worker"])-1
                evaluator_node = cluster["cluster"]["worker"][last_worker_index]
                cluster["cluster"]["evaluator"] = [evaluator_node]
                del cluster["cluster"]["worker"][last_worker_index]
                if evaluator_node == host_port:
                    cluster["task"] = {"type": "evaluator", "index": 0}

            print('TF_CONFIG: {} '.format(cluster))

            if num_executors > 1:
                os.environ["TF_CONFIG"] = json.dumps(cluster)

            is_chief = (cluster["task"]["type"] == "chief")

            logfile = experiment_utils._init_logger(experiment_utils._get_logdir(app_id, run_id), role=cluster["task"]["type"], index=cluster["task"]["index"])

            dist_logdir = experiment_utils._get_logdir(app_id, run_id) + '/logdir'
            if is_chief:
                hdfs.mkdir(dist_logdir)
                tensorboard._register(dist_logdir, experiment_utils._get_logdir(app_id, run_id), executor_num, local_logdir=local_logdir)
            else:
                tensorboard.events_logdir = dist_logdir

            print(devices._get_gpu_info())
            print('-------------------------------------------------------')
            print('Started running task')
            task_start = time.time()
            retval = train_fn()

            if is_chief:
                experiment_utils._handle_return_simple(retval, experiment_utils._get_logdir(app_id, run_id), logfile)

            task_end = time.time()
            time_str = 'Finished task - took ' + experiment_utils._time_diff(task_start, task_end)
            print(time_str)
            print('-------------------------------------------------------')
        except:
            raise
        finally:
            experiment_utils._cleanup(tensorboard, t)

    return _wrapper_fun


"""
Common Exceptions thrown by the hops library
"""

class RestAPIError(Exception):
    """This exception will be raised if there is an error response from a REST API call to Hopsworks"""

class UnkownSecretStorageError(Exception):
    """This exception will be raised if an unused secrets storage is passed as a parameter"""

class APIKeyFileNotFound(Exception):
    """This exception will be raised if the file with the API Key is not found in the local filesystem"""


"""
API for reading/writing numpy arrays to/from HDFS
"""
import hops.hdfs as hdfs
import numpy as np
import os

def load(hdfs_filename, **kwds):
    """
    Reads a file from HDFS into a Numpy Array

     Args:
       :hdfs_filename: You can specify either a full hdfs pathname or a relative one (relative to your Project's path in HDFS).
       :**kwds: You can add any additional args found in numpy.load(...) 

     Returns:
      A numpy array

     Raises:
      IOError: If the file does not exist
    """
    local_path = _copyHdfsToLocalOverwrite(hdfs_filename)
    return np.load(local_path, **kwds)


def loadtxt(hdfs_filename, **kwds):
    """
    Load data from a text file in HDFS into a Numpy Array. 
    Each row in the text file must have the same number of values.

     Args:
       :hdfs_filename: You can specify either a full hdfs pathname or a relative one (relative to your Project's path in HDFS).
       :**kwds: You can add any additional args found in numpy.loadtxt(...) 

     Returns:
      A numpy array

     Raises:
      IOError: If the file does not exist
    """
    local_path = _copyHdfsToLocalOverwrite(hdfs_filename)
    return np.loadtxt(local_path, **kwds)

def genfromtxt(hdfs_filename, **kwds):
    """
    Load data from a HDFS text file, with missing values handled as specified.
    Each line past the first skip_header lines is split at the delimiter character, and characters following the comments character are discarded.

     Args:
       :hdfs_filename: You can specify either a full hdfs pathname or a relative one (relative to your Project's path in HDFS).
       :**kwds: You can add any additional args found in numpy.loadtxt(...) 

     Returns:
      A numpy array

     Raises:
      IOError: If the file does not exist
    """
    local_path = _copyHdfsToLocalOverwrite(hdfs_filename)
    return np.genfromtxt(local_path, **kwds)

def fromregex(hdfs_filename, **kwds):
    """
    Construct an array from a text file, using regular expression parsing.

     Args:
       :hdfs_filename: You can specify either a full hdfs pathname or a relative one (relative to your Project's path in HDFS).
       :**kwds: You can add any additional args found in numpy.loadtxt(...) 

     Returns:
      A numpy array

     Raises:
      IOError: If the file does not exist
    """
    local_path = _copyHdfsToLocalOverwrite(hdfs_filename)
    return np.fromregex(local_path, **kwds)

def fromfile(hdfs_filename, **kwds):
    """
    Construct an array from data in a text or binary file.

     Args:
       :hdfs_filename: You can specify either a full hdfs pathname or a relative one (relative to your Project's path in HDFS).
       :**kwds: You can add any additional args found in numpy.loadtxt(...) 

     Returns:
      A numpy array

     Raises:
      IOError: If the file does not exist
    """
    local_path = _copyHdfsToLocalOverwrite(hdfs_filename)
    return np.fromregex(local_path, **kwds)


def memmap(hdfs_filename, **kwds):
    """
    Create a memory-map to an array stored in a binary file on disk.

     Args:
       :hdfs_filename: You can specify either a full hdfs pathname or a relative one (relative to your Project's path in HDFS).
       :**kwds: You can add any additional args found in numpy.loadtxt(...) 

     Returns:
      A memmap with dtype and shape that matches the data.

     Raises:
      IOError: If the file does not exist
    """
    local_path = _copyHdfsToLocalOverwrite(hdfs_filename)
    return np.fromregex(local_path, **kwds)


def save(hdfs_filename, data):
    """
    Saves a numpy array to a file in HDFS

    Args:
       :hdfs_filename: You can specify either a full hdfs pathname or a relative one (relative to your Project's path in HDFS)
       :data: numpy array

    Raises:
      IOError: If the local file does not exist
    """
    local_file = os.path.basename(hdfs_filename)
    np.save(local_file, data)
    _copyToHdfsOverwrite(hdfs_filename)

def savez(hdfs_filename, *args, **kwds):
    """
    Save several arrays into a single file in uncompressed .npz format in HDFS
    If arguments are passed in with no keywords, the corresponding variable names, in the .npz file, are 'arr_0', 'arr_1', etc.
    If keyword arguments are given, the corresponding variable names, in the .npz file will match the keyword names.

    Args:
       :hdfs_filename: You can specify either a full hdfs pathname or a relative one (relative to your Project's path in HDFS)
       :args: Arguments, optional
               Arrays to save to the file. Since it is not possible for Python to know the names of the arrays outside savez, 
               the arrays will be saved with names 'arr_0', 'arr_1', and so on. These arguments can be any expression.
       :kwds: Keyword arguments, optional
               Arrays to save to the file. Arrays will be saved in the file with the keyword names.       :data: numpy array

    Returns: None

    Raises:
      IOError: If the local file does not exist
    """
    local_file = os.path.basename(hdfs_filename)
    np.savez(local_file, *args, **kwds)
    _copyToHdfsOverwrite(hdfs_filename)

def savez_compressed(hdfs_filename, *args, **kwds):
    """
    Save several arrays into a single file in uncompressed .npz format in HDFS
    If arguments are passed in with no keywords, the corresponding variable names, in the .npz file, are 'arr_0', 'arr_1', etc.
    If keyword arguments are given, the corresponding variable names, in the .npz file will match the keyword names.

    Args:
       :hdfs_filename: You can specify either a full hdfs pathname or a relative one (relative to your Project's path in HDFS)
       :*args: Arguments, optional
               Arrays to save to the file. Since it is not possible for Python to know the names of the arrays outside savez, 
               the arrays will be saved with names 'arr_0', 'arr_1', and so on. These arguments can be any expression.
       :**kwds: Keyword arguments, optional
               Arrays to save to the file. Arrays will be saved in the file with the keyword names.       :data: numpy array

    Returns: None

    Raises:
      IOError: If the local file does not exist
    """
    local_file = os.path.basename(hdfs_filename)
    np.savez_compressed(local_file, *args, **kwds)
    _copyToHdfsOverwrite(hdfs_filename)


def _copyHdfsToLocalOverwrite(hdfs_filename):
    hdfs_path = hdfs._expand_path(hdfs_filename)
    local_path = hdfs.copy_to_local(hdfs_path, overwrite=True)
    return local_path
    
def _copyToHdfsOverwrite(hdfs_filename):
    local_file = os.path.basename(hdfs_filename)
    hdfs_path = hdfs._expand_path(hdfs_filename, exists=False)
    if local_file in hdfs_path:
        # copy_to_hdfs expects directory to copy to, excluding the file name
        hdfs_path = hdfs_path.replace(local_file, "")
    hdfs.copy_to_hdfs(local_file, hdfs_path, overwrite=True)



"""

Utility functions to manage users in Hopsworks.

"""
import json
from hops import constants, util

import logging
log = logging.getLogger(__name__)

def create_user(new_user):
    """
    Create a user in Hopsworks. Registers and activates a user with role HOPS_USER.

    Example usage:

    >>> from hops import util, user
    >>> new_user = {"firstName":"Joe","lastName":"Doe","email":"joe@hopsworks.ai","telephoneNum":"",
    >>>             "chosenPassword":"Admin123","repeatedPassword":"Admin123",
    >>>             "securityQuestion":"What is your oldest sibling's middle name?","securityAnswer":"Admin123",
    >>>             "tos":"true","authType":"Mobile","twoFactor":"false","toursEnabled":"true","orgName":"","dep":"",
    >>>             "street":"","city":"","postCode":"","country":"","testUser":"false"}
    >>> util.connect("localhost", api_key="api_key_file")
    >>> user.create(new_user)

    Args:
        :new_user: Dict with the new user attributes

    Returns:
        None
    """
    headers = {constants.HTTP_CONFIG.HTTP_CONTENT_TYPE: constants.HTTP_CONFIG.HTTP_APPLICATION_JSON}
    # Register user
    util.http(constants.DELIMITERS.SLASH_DELIMITER + \
              constants.REST_CONFIG.HOPSWORKS_REST_RESOURCE + constants.DELIMITERS.SLASH_DELIMITER + \
              constants.REST_CONFIG.HOPSWORKS_AUTH_RESOURCE + constants.DELIMITERS.SLASH_DELIMITER + \
              constants.REST_CONFIG.HOPSWORKS_AUTH_RESOURCE_REGISTER,
              headers=headers,
              method=constants.HTTP_CONFIG.HTTP_POST,
              data=json.dumps(new_user))

    # Get user id
    response = util.http(constants.DELIMITERS.SLASH_DELIMITER + \
                         constants.REST_CONFIG.HOPSWORKS_REST_RESOURCE + constants.DELIMITERS.SLASH_DELIMITER + \
                         constants.REST_CONFIG.HOPSWORKS_ADMIN_RESOURCE + constants.DELIMITERS.SLASH_DELIMITER + \
                         constants.REST_CONFIG.HOPSWORKS_USERS_RESOURCE + "?filter_by=user_email:" + new_user["email"],
                         headers=headers,
                         method=constants.HTTP_CONFIG.HTTP_GET)
    user_profile = response['items'][0]
    user_profile["status"] = "VERIFIED_ACCOUNT"
    # Verify user
    util.http(constants.DELIMITERS.SLASH_DELIMITER + \
              constants.REST_CONFIG.HOPSWORKS_REST_RESOURCE + constants.DELIMITERS.SLASH_DELIMITER + \
              constants.REST_CONFIG.HOPSWORKS_ADMIN_RESOURCE + constants.DELIMITERS.SLASH_DELIMITER + \
              constants.REST_CONFIG.HOPSWORKS_USERS_RESOURCE + constants.DELIMITERS.SLASH_DELIMITER + \
              str(user_profile['id']) + constants.DELIMITERS.SLASH_DELIMITER,
              headers=headers,
              method=constants.HTTP_CONFIG.HTTP_PUT,
              data=json.dumps(user_profile))
    # Accept user
    response = util.http(constants.DELIMITERS.SLASH_DELIMITER + \
                         constants.REST_CONFIG.HOPSWORKS_REST_RESOURCE + constants.DELIMITERS.SLASH_DELIMITER + \
                         constants.REST_CONFIG.HOPSWORKS_ADMIN_RESOURCE + constants.DELIMITERS.SLASH_DELIMITER + \
                         constants.REST_CONFIG.HOPSWORKS_USERS_RESOURCE + constants.DELIMITERS.SLASH_DELIMITER + \
                         str(user_profile['id']) + constants.DELIMITERS.SLASH_DELIMITER + "accepted",
                         headers=headers,
                         method=constants.HTTP_CONFIG.HTTP_PUT,
                         data=json.dumps(user_profile))
    log.info(response)



"""
API for opening csv files into Pandas from HDFS
"""
import hops.hdfs as hdfs
import pandas as pd

def read_csv(hdfs_filename, **kwds):
    """
      Reads a comma-separated values (csv) file from HDFS into a Pandas DataFrame

      Args:
         :hdfs_filename: You can specify either a full hdfs pathname or a relative one (relative to your Project's path in HDFS)
         :**kwds: You can add any additional args found in pandas.read_csv(...) 

      Returns:
        A pandas dataframe

      Raises:
        IOError: If the file does not exist
    """
    hdfs_path = hdfs._expand_path(hdfs_filename)    
    h = hdfs.get_fs()
    with h.open_file(hdfs_path, "rt") as f:
      data = pd.read_csv(f, **kwds)
    return data

def read_parquet(hdfs_filename, **kwds):
    """
      Load a parquet object from a HDFS path, returning a DataFrame.

      Args:
         :hdfs_filename: You can specify either a full hdfs pathname or a relative one (relative to your Project's path in HDFS)
         :**kwds: You can add any additional args found in pandas.read_csv(...) 

      Returns:
        A pandas dataframe

      Raises:
        IOError: If the file does not exist
    """
    hdfs_path = hdfs._expand_path(hdfs_filename)    
    h = hdfs.get_fs()
    with h.open_file(hdfs_path, "rt") as f:
      data = pd.read_parquet(f, **kwds)
    return data

def read_json(hdfs_filename, **kwds):
    """
      Convert a JSON string to pandas object.

      Args:
         :hdfs_filename: You can specify either a full hdfs pathname or a relative one (relative to your Project's path in HDFS)
         :**kwds: You can add any additional args found in pandas.read_csv(...) 

      Returns:
        A pandas dataframe

      Raises:
        IOError: If the file does not exist
    """
    hdfs_path = hdfs._expand_path(hdfs_filename)    
    h = hdfs.get_fs()
    with h.open_file(hdfs_path, "rt") as f:
      data = pd.read_json(f, **kwds)
    return data

def read_excel(hdfs_filename, **kwds):
    """
      Retrieve pandas object stored in HDFS file, optionally based on where criteria



      Args:
         :hdfs_filename: You can specify either a full hdfs pathname or a relative one (relative to your Project's path in HDFS)
         :**kwds: You can add any additional args found in pandas.read_csv(...) 

      Returns:
        A pandas dataframe

      Raises:
        IOError: If the file does not exist
    """
    hdfs_path = hdfs._expand_path(hdfs_filename)    
    h = hdfs.get_fs()
    with h.open_file(hdfs_path, "rt") as f:
      data = pd.read_excel(f, **kwds)
    return data


def write_csv(hdfs_filename, dataframe, **kwds):
    """
      Writes a pandas dataframe to a comma-separated values (csv) text file in HDFS. Overwrites the file if it already exists

      Args:
         :hdfs_filename: You can specify either a full hdfs pathname or a relative one (relative to your Project's path in HDFS)
         :dataframe: a Pandas dataframe
         :**kwds: You can add any additional args found in pandas.to_csv(...) 

      Raises:
        IOError: If the file does not exist
    """
    hdfs_path = hdfs._expand_path(hdfs_filename, exists=False)    
    h = hdfs.get_fs()
    with h.open_file(hdfs_path, "wt") as f:
      dataframe.to_csv(f, **kwds)


def write_parquet(hdfs_filename, dataframe, **kwds):
    """
      Writes a pandas dataframe to a parquet file in HDFS. Overwrites the file if it already exists

      Args:
         :hdfs_filename: You can specify either a full hdfs pathname or a relative one (relative to your Project's path in HDFS)
         :dataframe: a Pandas dataframe
         :**kwds: You can add any additional args found in pandas.to_parequet(...) 

      Raises:
        IOError: If the file does not exist
    """
    hdfs_path = hdfs._expand_path(hdfs_filename, exists=False)    
    h = hdfs.get_fs()
    with h.open_file(hdfs_path, "wb") as f:
      dataframe.to_parquet(f, **kwds)


def write_json(hdfs_filename, dataframe, **kwds):
    """
      Writes a pandas dataframe to a JSON file in HDFS. Overwrites the file if it already exists

      Args:
         :hdfs_filename: You can specify either a full hdfs pathname or a relative one (relative to your Project's path in HDFS)
         :dataframe: a Pandas dataframe
         :**kwds: You can add any additional args found in pandas.to_json(...) 

      Raises:
        IOError: If the file does not exist
    """
    hdfs_path = hdfs._expand_path(hdfs_filename, exists=False)    
    h = hdfs.get_fs()
    with h.open_file(hdfs_path, "wt") as f:
      dataframe.to_json(f, **kwds)

      



"""

Utility functions to manage the lifecycle of TensorBoard and get the path to write TensorBoard events.

"""
import socket
import subprocess
import time
import os
from hops import hdfs as hopshdfs
from hops.experiment_impl.util import experiment_utils
from hops import util
import shutil

root_logdir_path = None
events_logdir = None
tb_pid = 0
tb_url = None
tb_port = None
endpoint = None
debugger_endpoint = None
pypath = None
tb_path = None
local_logdir_path = None
local_logdir_bool = False

def _register(hdfs_exec_dir, endpoint_dir, exec_num, local_logdir=False):
    """

    Args:
        hdfs_exec_dir:
        endpoint_dir:
        exec_num:
        local_logdir:

    Returns:

    """
    global tb_pid

    if tb_pid != 0:
        subprocess.Popen(["kill", str(tb_pid)])

    _reset_global()

    global events_logdir
    events_logdir = hdfs_exec_dir

    global local_logdir_bool
    local_logdir_bool = local_logdir


    if tb_pid == 0:
        global pypath
        pypath = os.getenv("PYSPARK_PYTHON")

        #find free port
        tb_socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        tb_socket.bind(('',0))
        global tb_port
        tb_addr, tb_port = tb_socket.getsockname()

        global tb_path
        tb_path = experiment_utils._find_tensorboard()

        tb_socket.close()

        tb_env = _init_tb_env()

        global local_logdir_path
        if local_logdir:
            local_logdir_path = os.getcwd() + '/local_logdir'
            if os.path.exists(local_logdir_path):
                shutil.rmtree(local_logdir_path)
                os.makedirs(local_logdir_path)
            else:
                os.makedirs(local_logdir_path)

            local_logdir_path = local_logdir_path + '/'
            tb_proc = subprocess.Popen([pypath, tb_path, "--logdir=%s" % local_logdir_path, "--port=%d" % tb_port, "--host=%s" % "0.0.0.0"],
                                       env=tb_env, preexec_fn=util._on_executor_exit('SIGTERM'))
        else:
            tb_proc = subprocess.Popen([pypath, tb_path, "--logdir=%s" % events_logdir, "--port=%d" % tb_port, "--host=%s" % "0.0.0.0"],
                                   env=tb_env, preexec_fn=util._on_executor_exit('SIGTERM'))

        tb_pid = tb_proc.pid

        host = socket.getfqdn()
        global tb_url
        tb_url = "http://{0}:{1}".format(host, tb_port)
        global endpoint
        endpoint = endpoint_dir + "/TensorBoard.task" + str(exec_num)

        #dump tb host:port to hdfs
    hopshdfs.dump(tb_url, endpoint)

    return endpoint, tb_pid

def logdir():
    """
    Get the TensorBoard logdir. This function should be called in your wrapper function for Experiment, Parallel Experiment or Distributed Training and passed as the
    logdir for TensorBoard.

    *Case 1: local_logdir=True*, then the logdir is on the local filesystem, otherwise it is in the folder for your experiment in your project in HDFS. Once the experiment is finished all the files that are present in the directory will be uploaded to tour experiment directory in the Experiments dataset.

    *Case 2: local_logdir=False*, then the logdir is in HDFS in your experiment directory in the Experiments dataset.

    Returns:
        The path to store files for your experiment. The content is also visualized in TensorBoard.
    """

    global local_logdir_bool
    if local_logdir_bool:
        return local_logdir_path

    global events_logdir
    return events_logdir

def interactive_debugger():
    """

    Returns: address for interactive debugger in TensorBoard


    """
    global debugger_endpoint
    debugger_endpoint =_restart_debugging()
    return debugger_endpoint

def non_interactive_debugger():
    """

    Returns: address for non-interactive debugger in TensorBoard

    """
    global debugger_endpoint
    debugger_endpoint =_restart_debugging(interactive=False)
    return debugger_endpoint

def _restart_debugging(interactive=True):
    """

    Args:
        interactive:

    Returns:

    """
    global tb_pid

    #Kill existing TB
    proc = subprocess.Popen(["kill", str(tb_pid)])
    proc.wait()

    debugger_socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    debugger_socket.bind(('',0))
    debugger_addr, debugger_port = debugger_socket.getfqdn()

    debugger_socket.close()

    tb_env = _init_tb_env()

    global pypath
    global tb_path
    global tb_port

    if interactive:
        tb_proc = subprocess.Popen([pypath, tb_path, "--logdir=%s" % logdir(), "--port=%d" % tb_port, "--debugger_port=%d" % debugger_port, "--host=%s" % "0.0.0.0"],
                                   env=tb_env, preexec_fn=util._on_executor_exit('SIGTERM'))
        tb_pid = tb_proc.pid

    if not interactive:
        tb_proc = subprocess.Popen([pypath, tb_path, "--logdir=%s" % logdir(), "--port=%d" % tb_port, "--debugger_data_server_grpc_port=%d" % debugger_port, "--host=%s" % "0.0.0.0"],
                                   env=tb_env, preexec_fn=util._on_executor_exit('SIGTERM'))
        tb_pid = tb_proc.pid

    time.sleep(2)

    return 'localhost:' + str(debugger_port)

def _init_tb_env():
    tb_env = os.environ.copy()
    tb_env['CUDA_VISIBLE_DEVICES'] = ''
    tb_env['HIP_VISIBLE_DEVICES'] = '-1'
    tb_env['LC_ALL'] = 'C'
    tb_env['TMPDIR'] = os.getcwd()
    return tb_env


def _reset_global():
    """

    Returns:

    """
    global root_logdir_path
    global events_logdir
    global tb_pid
    global tb_url
    global tb_port
    global endpoint
    global debugger_endpoint
    global pypath
    global tb_path
    global local_logdir_path
    global local_logdir_bool
    
    root_logdir_path = None
    events_logdir = None
    tb_pid = 0
    tb_url = None
    tb_port = None
    endpoint = None
    debugger_endpoint = None
    pypath = None
    tb_path = None
    local_logdir_path = None
    local_logdir_bool = False



"""
A module for working with Hopsworks datasets.
"""

import re
import os
import ntpath
import math
import json
import time

from hops import constants, util, project
from hops.exceptions import RestAPIError

import logging
log = logging.getLogger(__name__)

class HTTPUpload:
    size_last_chunk = 0
    DEFAULT_FLOW_STANDARD_CHUNK_SIZE = 1048576
    flow_standard_chunk_size = DEFAULT_FLOW_STANDARD_CHUNK_SIZE
    resource = None
    file = None
    f = None
    params = {}

    def __init__(self, file, path, flow_standard_chunk_size=DEFAULT_FLOW_STANDARD_CHUNK_SIZE):
        self.resource = constants.REST_CONFIG.HOPSWORKS_REST_RESOURCE + \
                        constants.DELIMITERS.SLASH_DELIMITER + \
                        constants.REST_CONFIG.HOPSWORKS_PROJECT_RESOURCE + \
                        constants.DELIMITERS.SLASH_DELIMITER + \
                        os.environ[constants.ENV_VARIABLES.HOPSWORKS_PROJECT_ID_ENV_VAR] + \
                        constants.DELIMITERS.SLASH_DELIMITER + \
                        constants.REST_CONFIG.HOPSWORKS_DATASETS_RESOURCE + \
                        constants.DELIMITERS.SLASH_DELIMITER + \
                        "upload" + \
                        constants.DELIMITERS.SLASH_DELIMITER + \
                        path

        self.file = file
        if not flow_standard_chunk_size:
            self.flow_standard_chunk_size = self.DEFAULT_FLOW_STANDARD_CHUNK_SIZE
        else:
            self.flow_standard_chunk_size = flow_standard_chunk_size
        self._add_flow_params()

    def _add_flow_params(self):
        self.params['templateId'] = -1
        self.params['flowChunkNumber'] = 1
        self.params['flowChunkSize'] = self.flow_standard_chunk_size
        self.params['flowTotalSize'] = self._calculate_file_size(self.file)
        self.params['flowCurrentChunkSize'] = self.flow_standard_chunk_size
        self.size_last_chunk = self.params['flowTotalSize'] % self.flow_standard_chunk_size
        self.params['flowIdentifier'] = str(self.params['flowTotalSize']) \
            + '-' \
            + re.sub("[^0-9A-Za-z_-]", "", self.file)
        self.params['flowFilename'] = self._path_leaf(self.file)
        self.params['flowRelativePath'] = self.file
        self.params['flowTotalChunks'] = self._calculate_total_chunks(self.flow_standard_chunk_size,
                                                                      self.params['flowTotalSize'])

    @staticmethod
    def _calculate_file_size(file):
        return os.path.getsize(file)

    @staticmethod
    def _calculate_size_last_chunk(chunk_size, file_size):
        return file_size % chunk_size

    @staticmethod
    def _calculate_total_chunks(chunk_size, file_size):
        chunks = math.ceil(file_size / chunk_size)
        if chunks == 0:
            chunks = 1
        return chunks

    def _next(self):
        if self.params['flowChunkNumber'] > self.params['flowTotalChunks']:
            raise Exception

        chunk_size = self.params['flowChunkSize']
        if self.params['flowChunkNumber'] == self.params['flowTotalChunks']:
            if chunk_size >= self.size_last_chunk:
                chunk_size = self.size_last_chunk
            else:
                chunk_size += self.size_last_chunk
        if self.params['flowTotalSize'] < self.params['flowChunkSize']:
            chunk_size = self.size_last_chunk

        chunk = self._read_chunk(chunk_size)
        self.params['flowCurrentChunkSize'] = chunk_size

        # Upload chunk
        response = util.send_request(constants.HTTP_CONFIG.HTTP_POST,
                                     resource="/" + self.resource,
                                     data=self.params,
                                     files={'file': (self.file, chunk)})

        response_object = response.json()
        if response.status_code >= 400:
            error_code, error_msg, user_msg = util._parse_rest_error(response_object)
            raise RestAPIError("Could not perform action on job's execution (url: {}), server response: \n "
                               "HTTP code: {}, HTTP reason: {}, error code: {}, error msg: {}, user msg: {}".format(
                                self.resource, response.status_code, response.reason, error_code, error_msg, user_msg))

        progress = round(self.params['flowChunkNumber'] / self.params['flowTotalChunks'] * 100, 3)
        log.info("Progress: " + str(progress) + "%")
        self.params['flowChunkNumber'] += 1

    def _read_chunk(self, chunk_size=1024):
        data = self.f.read(chunk_size)
        return data

    @staticmethod
    def _path_leaf(path):
        head, tail = ntpath.split(path)
        return tail or ntpath.basename(head)

    def upload(self):
        with open(self.file, "rb") as self.f:
            while self.params['flowChunkNumber'] <= self.params['flowTotalChunks']:
                self._next()


class HTTPDownload:
    # stream=True and chunk_size=None will read data as it arrives in whatever size the chunks are received
    DEFAULT_DOWNLOAD_CHUNK_SIZE = None

    def __init__(self, path, file, chunk_size=DEFAULT_DOWNLOAD_CHUNK_SIZE):
        self.resource = constants.REST_CONFIG.HOPSWORKS_REST_RESOURCE + \
                        constants.DELIMITERS.SLASH_DELIMITER + \
                        constants.REST_CONFIG.HOPSWORKS_PROJECT_RESOURCE + \
                        constants.DELIMITERS.SLASH_DELIMITER + \
                        os.environ[constants.ENV_VARIABLES.HOPSWORKS_PROJECT_ID_ENV_VAR] + \
                        constants.DELIMITERS.SLASH_DELIMITER + \
                        constants.REST_CONFIG.HOPSWORKS_DATASETS_RESOURCE + \
                        constants.DELIMITERS.SLASH_DELIMITER + \
                        "download" + \
                        constants.DELIMITERS.SLASH_DELIMITER + \
                        "with_auth" + \
                        constants.DELIMITERS.SLASH_DELIMITER + \
                        path + \
                        constants.DELIMITERS.QUESTION_MARK_DELIMITER + \
                        "type=DATASET"

        self.file = file
        if not chunk_size:
            self.chunk_size = self.DEFAULT_DOWNLOAD_CHUNK_SIZE
        else:
            self.chunk_size = chunk_size

    def download(self):
        with util.send_request(constants.HTTP_CONFIG.HTTP_GET,
                               resource="/" + self.resource,
                               stream=True) as response:

            if response.status_code // 100 != 2:
                error_code, error_msg, user_msg = util._parse_rest_error(response.json())
                raise RestAPIError("Could not perform action on job's execution (url: {}), server response: \n "
                                   "HTTP code: {}, HTTP reason: {}, error code: {}, error msg: {}, user msg: {}".format(
                                    self.resource, response.status_code, response.reason, error_code, error_msg, user_msg))

            with open(self.file, "wb") as f:
                downloaded = 0
                file_size = response.headers.get('Content-Length')
                if not file_size:
                    log.debug("Downloading file ...", end=" ")
                for chunk in response.iter_content(chunk_size=self.chunk_size):
                    f.write(chunk)
                    downloaded += len(chunk)
                    if file_size:
                        progress = round(downloaded / int(file_size) * 100, 3)
                        log.info("Progress: " + str(progress) + "%")
                if not file_size:
                    log.debug("DONE")
    

def upload(file, remote_path, chunk_size=None):
    """
    Upload data to a project's dataset by setting the path to the local file to be uploaded to a remote_path in a
    dataset. The file is split into chunks which are uploaded sequentially.
    If you run this method more than once for the same file and remote_path, if the file already exists in Hopsworks
    it will be overwritten.

    Example usage:

    >>> from hops import dataset
    >>> dataset.upload("/tmp/mylocalfile.txt", "Resources/uploaded_files_dir")

    Args:
        :file: the absolute path or the filename of the file to be uploaded.
        :remote_path: the dataset or the path to the folder in the dataset to upload the file.
        :chunk_size: (Optional) chunk size used to upload the file

    Returns:
        None
    """
    HTTPUpload(file, remote_path, chunk_size).upload()


def download(remote_path, file, chunk_size=None):
    """
    Download a file from a project's dataset by specifying the path to the remote file and a local path where the
    file is downloaded. The remote file path must include the project name as shown in the example. It is possible
    to download files from shared projects by setting the shared project name. If the chunk size is specified, the remote
    file is downloaded in chunks of similar size. Otherwise, chunks can have any size and are read as they are received.

    Example usage:

    >>> from hops import dataset
    >>> dataset.download("Projects/project_name/Resources/myremotefile.txt", "tmp/mylocalfile.txt")

    Args:
        :remote_path: the path to the remote file in the dataset
        :file: the absolute or relative path where the file is downloaded
        :chunk_size: (Optional) chunk size used to download the file

    Returns:
        None
    """
    HTTPDownload(remote_path, file, chunk_size).download()


def get_path_info(remote_path, project_name=None):
    """
    Check if file exists.

    Example usage:

    >>> from hops import dataset
    >>> dataset.get_path_info("Projects/project_name/Resources/myremotefile.txt")

    Args:
        :remote_path: the path to the remote file or directory in the dataset.
        :project_name: whether this method should wait for the zipping process to complete beefore returning.

    Returns:
        A json representation of the path metadata.
    """
    project_id = project.project_id_as_shared(project_name)
    resource_url = constants.DELIMITERS.SLASH_DELIMITER + \
                   constants.REST_CONFIG.HOPSWORKS_REST_RESOURCE + constants.DELIMITERS.SLASH_DELIMITER + \
                   constants.REST_CONFIG.HOPSWORKS_PROJECT_RESOURCE + constants.DELIMITERS.SLASH_DELIMITER + \
                   project_id + constants.DELIMITERS.SLASH_DELIMITER + \
                   constants.REST_CONFIG.HOPSWORKS_DATASETS_RESOURCE + constants.DELIMITERS.SLASH_DELIMITER + \
                   remote_path + "?action=stat"

    response = util.send_request('GET', resource_url)
    response_object = response.json()

    if response.status_code >= 400:
        error_code, error_msg, user_msg = util._parse_rest_error(response_object)
        raise RestAPIError("Could not get path (url: {}), server response: \n "
                           "HTTP code: {}, HTTP reason: {}, error code: {}, error msg: {}, user msg: {}".format(
            resource_url, response.status_code, response.reason, error_code, error_msg, user_msg))
    else:
        return json.loads(response.content)


def path_exists(remote_path, project_name=None):
    """
    Check if path exists.

    Example usage:

    >>> from hops import dataset
    >>> dataset.path_exists("Projects/project_name/Resources/myremotefile.txt")

    Args:
        :remote_path: the path to the remote file or directory in the dataset
        :project_name: whether this method should wait for the zipping process to complete beefore returning.

    Returns:
        True if path exists, False otherwise
    """
    try:
        get_path_info(remote_path, project_name)
        return True
    except RestAPIError:
        return False


def delete(remote_path, project_name=None, block=True, timeout=30):
    """
    Delete the dir or file in Hopsworks, specified by the remote_path.

    Example usage:

    >>> from hops import dataset
    >>> dataset.delete("Projects/project_name/Resources/myremotefile.txt")

    Args:
        :remote_path: the path to the remote file or directory in the dataset
        :project_name: whether this method should wait for the zipping process to complete before returning.
        :block: whether to wait for the deletion to complete or not.
        :timeout: number of seconds to wait for the deletion to complete before returning.

    Returns:
        None
    """
    project_id = project.project_id_as_shared(project_name)
    resource_url = constants.DELIMITERS.SLASH_DELIMITER + \
                   constants.REST_CONFIG.HOPSWORKS_REST_RESOURCE + constants.DELIMITERS.SLASH_DELIMITER + \
                   constants.REST_CONFIG.HOPSWORKS_PROJECT_RESOURCE + constants.DELIMITERS.SLASH_DELIMITER + \
                   project_id + constants.DELIMITERS.SLASH_DELIMITER + \
                   constants.REST_CONFIG.HOPSWORKS_DATASETS_RESOURCE + constants.DELIMITERS.SLASH_DELIMITER + \
                   remote_path

    util.send_request('DELETE', resource_url)

    # Check if path is indeed deleted as REST API is asynchronous
    if block:
        count = 0
        while count < timeout and path_exists(remote_path, project_name):
            log.debug("Waiting for deletion...")
            count += 1
            time.sleep(1)

        if count >= timeout:
            raise DeletionTimeout("Timeout of {} seconds exceeded while deleting path {}.".format(timeout, remote_path))


def compress(remote_path, project_name=None, block=False, timeout=120):
    """
    Compress the dir or file in Hopsworks, specified by the remote_path.

    Example usage:

    >>> from hops import dataset
    >>> dataset.compress("Projects/project_name/Resources/myremotefile.txt")

    Args:
        :remote_path: the path to the remote file or directory in the dataset
        :project_name: whether this method should wait for the zipping process to complete before returning.
        :block: whether to wait for the compression to complete or not.
        :timeout: number of seconds to wait for the compression to complete before returning.

    Returns:
        None
    """
    _archive(remote_path, project_name=project_name, block=block, timeout=timeout, action='zip')


def extract(remote_path, project_name=None, block=False, timeout=120):
    """
    Extract the dir or file in Hopsworks, specified by the remote_path.

    Example usage:

    >>> from hops import dataset
    >>> dataset.extract("Projects/project_name/Resources/myremotefile.zip")

    Args:
        :remote_path: the path to the remote file or directory in the dataset
        :project_name: whether this method should wait for the zipping process to complete before returning.
        :block: whether to wait for the extraction to complete or not.
        :timeout: number of seconds to wait for the extraction to complete before returning.

    Returns:
        None
    """
    _archive(remote_path, project_name=project_name, block=block, timeout=timeout, action='unzip')


def _archive(remote_path, project_name=None, block=False, timeout=120, action='zip'):
    """
    Create an archive (zip file) of a file or directory in a Hopsworks dataset.

    Args:
        :remote_path: the path to the remote file or directory in the dataset.
        :action: Allowed values are zip/unzip. Whether to compress/extract respectively.
        :block: whether this method should wait for the zipping process to complete before returning.
        :project_name: whether this method should wait for the zipping process to complete beefore returning.
        :timeout: number of seconds to wait for the action to complete before returning.
    Returns:
        None
    """
    project_id = project.project_id_as_shared(project_name)
    resource_url = constants.DELIMITERS.SLASH_DELIMITER + \
                   constants.REST_CONFIG.HOPSWORKS_REST_RESOURCE + constants.DELIMITERS.SLASH_DELIMITER + \
                   constants.REST_CONFIG.HOPSWORKS_PROJECT_RESOURCE + constants.DELIMITERS.SLASH_DELIMITER + \
                   project_id + constants.DELIMITERS.SLASH_DELIMITER + \
                   constants.REST_CONFIG.HOPSWORKS_DATASETS_RESOURCE + constants.DELIMITERS.SLASH_DELIMITER + \
                   remote_path + "?action=" + action

    util.send_request('POST', resource_url)

    if block is True:
        # Wait for zip file to appear. When it does, check that parent dir zipState is not set to CHOWNING
        count = 0
        while count < timeout:
            # Get the status of the zipped file
            zip_exists = path_exists(remote_path + ".zip", project_name)
            # Get the zipState of the directory being zipped
            dir_status = get_path_info(remote_path, project_name)
            zip_state = dir_status['zipState'] if 'zipState' in dir_status else None

            if zip_exists and zip_state == 'NONE' :
                log.info("Zipping completed.")
                return
            else:
                log.info("Zipping...")
                time.sleep(1)
            count += 1
        raise CompressTimeout("Timeout of {} seconds exceeded while compressing {}.".format(timeout, remote_path))


class CompressTimeout(Exception):
    """This exception will be raised if the compression of a path times out."""


class DeletionTimeout(Exception):
    """This exception will be raised if the deletion of a path in a dataset times out."""



"""
Simple experiment implementation
"""

from hops.experiment_impl.util import experiment_utils
from hops import devices, tensorboard, hdfs

import threading
import time
import json
import os
import six


def _run(sc, train_fn, run_id, args_dict=None, local_logdir=False, name="no-name"):
    """

    Args:
        sc:
        train_fn:
        args_dict:
        local_logdir:
        name:

    Returns:

    """

    app_id = str(sc.applicationId)


    if args_dict == None:
        num_executions = 1
    else:
        arg_lists = list(args_dict.values())
        currentLen = len(arg_lists[0])
        for i in range(len(arg_lists)):
            if currentLen != len(arg_lists[i]):
                raise ValueError('Length of each function argument list must be equal')
            num_executions = len(arg_lists[i])

    sc.setJobGroup(os.environ['ML_ID'], "{} | Launcher running experiment".format(name))
    #Each TF task should be run on 1 executor
    nodeRDD = sc.parallelize(range(num_executions), num_executions)

    #Force execution on executor, since GPU is located on executor
    nodeRDD.foreachPartition(_prepare_func(app_id, run_id, train_fn, args_dict, local_logdir))

    print('Finished Experiment \n')

    # For single run return .return if exists
    if args_dict == None:
        path_to_return = experiment_utils._get_logdir(app_id, run_id) + '/.outputs.json'
        if hdfs.exists(path_to_return):
            return_json = hdfs.load(path_to_return)
            return_dict = json.loads(return_json)
            return experiment_utils._get_logdir(app_id, run_id), return_dict
        else:
            return experiment_utils._get_logdir(app_id, run_id), None
    elif num_executions == 1:
        arg_count = six.get_function_code(train_fn).co_argcount
        arg_names = six.get_function_code(train_fn).co_varnames
        argIndex = 0
        param_string = ''
        while arg_count > 0:
            param_name = arg_names[argIndex]
            param_val = args_dict[param_name][0]
            param_string += str(param_name) + '=' + str(param_val) + '&'
            arg_count -= 1
            argIndex += 1
        param_string = param_string[:-1]
        path_to_return = experiment_utils._get_logdir(app_id, run_id) + '/' + param_string + '/.outputs.json'
        if hdfs.exists(path_to_return):
            return_json = hdfs.load(path_to_return)
            return_dict = json.loads(return_json)
            return experiment_utils._get_logdir(app_id, run_id), return_dict
        else:
            return experiment_utils._get_logdir(app_id, run_id), None
    else:
        return experiment_utils._get_logdir(app_id, run_id), None

#Helper to put Spark required parameter iter in function signature
def _prepare_func(app_id, run_id, train_fn, args_dict, local_logdir):
    """

    Args:
        app_id:
        run_id:
        train_fn:
        args_dict:
        local_logdir:

    Returns:

    """
    def _wrapper_fun(iter):
        """

        Args:
            iter:

        Returns:

        """

        for i in iter:
            executor_num = i

        experiment_utils._set_ml_id(app_id, run_id)

        tb_hdfs_path = ''

        hdfs_exec_logdir = experiment_utils._get_logdir(app_id, run_id)

        t = threading.Thread(target=devices._print_periodic_gpu_utilization)
        if devices.get_num_gpus() > 0:
            t.start()

        try:
            #Arguments
            if args_dict:
                param_string, params, args = experiment_utils.build_parameters(train_fn, executor_num, args_dict)
                hdfs_exec_logdir, hdfs_appid_logdir = experiment_utils._create_experiment_subdirectories(app_id, run_id, param_string, 'grid_search', params=params)
                logfile = experiment_utils._init_logger(hdfs_exec_logdir)
                tb_hdfs_path, tb_pid = tensorboard._register(hdfs_exec_logdir, hdfs_appid_logdir, executor_num, local_logdir=local_logdir)
                print(devices._get_gpu_info())
                print('-------------------------------------------------------')
                print('Started running task ' + param_string)
                task_start = time.time()
                retval = train_fn(*args)
                task_end = time.time()
                experiment_utils._handle_return_simple(retval, hdfs_exec_logdir, logfile)
                time_str = 'Finished task ' + param_string + ' - took ' + experiment_utils._time_diff(task_start, task_end)
                print(time_str)
                print('-------------------------------------------------------')
            else:
                tb_hdfs_path, tb_pid = tensorboard._register(hdfs_exec_logdir, hdfs_exec_logdir, executor_num, local_logdir=local_logdir)
                logfile = experiment_utils._init_logger(hdfs_exec_logdir)
                print(devices._get_gpu_info())
                print('-------------------------------------------------------')
                print('Started running task')
                task_start = time.time()
                retval = train_fn()
                task_end = time.time()
                experiment_utils._handle_return_simple(retval, hdfs_exec_logdir, logfile)
                time_str = 'Finished task - took ' + experiment_utils._time_diff(task_start, task_end)
                print(time_str)
                print('-------------------------------------------------------')
        except:
            raise
        finally:
            experiment_utils._cleanup(tensorboard, t)

    return _wrapper_fun